from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import lru_cache
from typing import Any

import pystac
from pystac.serialization import STACVersionRange
from pystac.serialization.identify import OldExtensionShortIDs, STACVersionID
from pystac.stac_object import STACObjectType


class SchemaUriMap(ABC):
    """Abstract class defining schema URIs for STAC core objects and extensions."""

    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_object_schema_uri(
        self, object_type: STACObjectType, stac_version: str
    ) -> str | None:
        """Get the schema URI for the given object type and stac version.

        Args:
            object_type : STAC object type. One of
                :class:`~pystac.STACObjectType`
            stac_version : The STAC version of the schema to return.

        Returns:
            str: The URI of the schema, or None if not found.
        """
        raise NotImplementedError


class DefaultSchemaUriMap(SchemaUriMap):
    """Implementation of SchemaUriMap that uses schemas hosted by stacspec.org

    For STAC Versions 0.9.0 or earlier this will use the schemas hosted on the
    radiantearth/stac-spec GitHub repo.
    """

    # BASE_URIS contains a list of tuples, the first element is a version range and the
    # second being the base URI for schemas for that range. The schema URI of a STAC
    # for a particular version uses the base URI associated with the version range which
    # contains it. If the version it outside of any VersionRange, there is no URI for
    # the schema.
    BASE_URIS: list[tuple[STACVersionRange, Callable[[str], str]]] = [
        (
            STACVersionRange(min_version="1.0.0-beta.2"),
            lambda version: f"https://schemas.stacspec.org/v{version}",
        ),
        (
            STACVersionRange(min_version="0.8.0", max_version="1.0.0-beta.1"),
            lambda version: (
                f"https://raw.githubusercontent.com/radiantearth/stac-spec/v{version}"
            ),
        ),
    ]

    # DEFAULT_SCHEMA_MAP contains a structure that matches 'core' or 'extension' schema
    # URIs based on the stac object type and the stac version, using a similar
    # technique as BASE_URIS. Uris are contained in a tuple whose first element
    # represents the URI of the latest version, so that a search through version
    # ranges is avoided if the STAC being validated
    # is the latest version. If it's a previous version, the stac_version that matches
    # the listed version range is used, or else the URI from the latest version is used
    # if there are no overrides for previous versions.
    DEFAULT_SCHEMA_MAP: dict[str, Any] = {
        STACObjectType.CATALOG: ("catalog-spec/json-schema/catalog.json", None),
        STACObjectType.COLLECTION: (
            "collection-spec/json-schema/collection.json",
            None,
        ),
        STACObjectType.ITEM: ("item-spec/json-schema/item.json", None),
    }

    @classmethod
    def _append_base_uri_if_needed(cls, uri: str, stac_version: str) -> str | None:
        # Only append the base URI if it's not already an absolute URI
        if "://" not in uri:
            for version_range, f in cls.BASE_URIS:
                if version_range.contains(stac_version):
                    base_uri = f(stac_version)
                    return f"{base_uri}/{uri}"

            # We don't have JSON schema validation for this version of PySTAC
            return None
        else:
            return uri

    def get_object_schema_uri(
        self, object_type: STACObjectType, stac_version: str
    ) -> str | None:
        is_latest = stac_version == pystac.get_stac_version()

        if object_type not in self.DEFAULT_SCHEMA_MAP:
            raise KeyError(f"Unknown STAC object type {object_type}")
        uri = self.DEFAULT_SCHEMA_MAP[object_type][0]
        if not is_latest:
            if self.DEFAULT_SCHEMA_MAP[object_type][1]:
                for version_range, range_uri in self.DEFAULT_SCHEMA_MAP[object_type][1]:
                    if version_range.contains(stac_version):
                        uri = range_uri
                        break

        return self._append_base_uri_if_needed(uri, stac_version)


class OldExtensionSchemaUriMap:
    """Ties old extension IDs to schemas hosted by https://schemas.stacspec.org.

    For STAC Versions 0.9.0 or earlier this will use the schemas hosted on the
    radiantearth/stac-spec GitHub repo.
    """

    # BASE_URIS contains a list of tuples, the first element is a version range and the
    # second being the base URI for schemas for that range. The schema URI of a STAC
    # for a particular version uses the base URI associated with the version range which
    # contains it. If the version it outside of any VersionRange, there is no URI for
    # the schema.
    @classmethod
    @lru_cache
    def get_base_uris(
        cls,
    ) -> list[tuple[STACVersionRange, Callable[[STACVersionID], str]]]:
        return [
            (
                STACVersionRange(min_version="1.0.0-beta.2"),
                lambda version: f"https://schemas.stacspec.org/v{version}",
            ),
            (
                STACVersionRange(min_version="0.8.0", max_version="1.0.0-beta.1"),
                lambda version: (
                    "https://raw.githubusercontent.com/"
                    f"radiantearth/stac-spec/v{version}"
                ),
            ),
        ]

    # DEFAULT_SCHEMA_MAP contains a structure that matches extension schema URIs
    # based on the stac object type, extension ID and the stac version.
    # Uris are contained in a tuple whose first element represents the URI of the latest
    # version, so that a search through version ranges is avoided if the STAC being
    # validated is the latest version. If it's a previous version, the stac_version
    # that matches the listed version range is used, or else the URI from the latest
    # version is used if there are no overrides for previous versions.
    @classmethod
    @lru_cache
    def get_schema_map(cls) -> dict[str, Any]:
        return {
            OldExtensionShortIDs.CHECKSUM.value: (
                {
                    pystac.STACObjectType.CATALOG: (
                        "extensions/checksum/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/checksum/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.ITEM: (
                        "extensions/checksum/json-schema/schema.json"
                    ),
                },
                None,
            ),
            OldExtensionShortIDs.COLLECTION_ASSETS.value: (
                {
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/collection-assets/json-schema/schema.json"
                    )
                },
                None,
            ),
            OldExtensionShortIDs.DATACUBE.value: (
                {
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/datacube/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.ITEM: (
                        "extensions/datacube/json-schema/schema.json"
                    ),
                },
                [
                    (
                        STACVersionRange(min_version="0.5.0", max_version="0.9.0"),
                        {
                            pystac.STACObjectType.COLLECTION: None,
                            pystac.STACObjectType.ITEM: None,
                        },
                    )
                ],
            ),
            OldExtensionShortIDs.EO.value: (
                {pystac.STACObjectType.ITEM: "extensions/eo/json-schema/schema.json"},
                None,
            ),
            OldExtensionShortIDs.ITEM_ASSETS.value: (
                {
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/item-assets/json-schema/schema.json"
                    )
                },
                None,
            ),
            OldExtensionShortIDs.LABEL.value: (
                {
                    pystac.STACObjectType.ITEM: (
                        "extensions/label/json-schema/schema.json"
                    )
                },
                [
                    (
                        STACVersionRange(min_version="0.8.0-rc1", max_version="0.8.1"),
                        {pystac.STACObjectType.ITEM: "extensions/label/schema.json"},
                    )
                ],
            ),
            OldExtensionShortIDs.POINTCLOUD.value: (
                # Invalid schema
                None,
                None,
            ),
            OldExtensionShortIDs.PROJECTION.value: (
                {
                    pystac.STACObjectType.ITEM: (
                        "extensions/projection/json-schema/schema.json"
                    )
                },
                None,
            ),
            OldExtensionShortIDs.SAR.value: (
                {pystac.STACObjectType.ITEM: "extensions/sar/json-schema/schema.json"},
                None,
            ),
            OldExtensionShortIDs.SAT.value: (
                {pystac.STACObjectType.ITEM: "extensions/sat/json-schema/schema.json"},
                None,
            ),
            OldExtensionShortIDs.SCIENTIFIC.value: (
                {
                    pystac.STACObjectType.ITEM: (
                        "extensions/scientific/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/scientific/json-schema/schema.json"
                    ),
                },
                None,
            ),
            OldExtensionShortIDs.SINGLE_FILE_STAC.value: (
                {
                    pystac.STACObjectType.CATALOG: (
                        "extensions/single-file-stac/json-schema/schema.json"
                    )
                },
                None,
            ),
            OldExtensionShortIDs.TILED_ASSETS.value: (
                {
                    pystac.STACObjectType.CATALOG: (
                        "extensions/tiled-assets/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/tiled-assets/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.ITEM: (
                        "extensions/tiled-assets/json-schema/schema.json"
                    ),
                },
                None,
            ),
            OldExtensionShortIDs.TIMESTAMPS.value: (
                {
                    pystac.STACObjectType.ITEM: (
                        "extensions/timestamps/json-schema/schema.json"
                    )
                },
                None,
            ),
            OldExtensionShortIDs.VERSION.value: (
                {
                    pystac.STACObjectType.ITEM: (
                        "extensions/version/json-schema/schema.json"
                    ),
                    pystac.STACObjectType.COLLECTION: (
                        "extensions/version/json-schema/schema.json"
                    ),
                },
                None,
            ),
            OldExtensionShortIDs.VIEW.value: (
                {pystac.STACObjectType.ITEM: "extensions/view/json-schema/schema.json"},
                None,
            ),
            # Removed or renamed extensions.
            "dtr": (None, None),  # Invalid schema
            "asset": (
                None,
                [
                    (
                        STACVersionRange(min_version="0.8.0-rc1", max_version="0.9.0"),
                        {
                            pystac.STACObjectType.COLLECTION: (
                                "extensions/asset/json-schema/schema.json"
                            )
                        },
                    )
                ],
            ),
        }

    @classmethod
    def _append_base_uri_if_needed(
        cls, uri: str, stac_version: STACVersionID
    ) -> str | None:
        # Only append the base URI if it's not already an absolute URI
        if "://" not in uri:
            for version_range, f in cls.get_base_uris():
                if version_range.contains(stac_version):
                    base_uri = f(stac_version)
                    return f"{base_uri}/{uri}"

            # No JSON Schema for the old extension
            return None
        else:
            return uri

    @classmethod
    def get_extension_schema_uri(
        cls, extension_id: str, object_type: STACObjectType, stac_version: STACVersionID
    ) -> str | None:
        uri = None

        is_latest = stac_version == pystac.get_stac_version()

        ext_map = cls.get_schema_map()
        if extension_id in ext_map:
            if ext_map[extension_id][0] and object_type in ext_map[extension_id][0]:
                uri = ext_map[extension_id][0][object_type]

            if not is_latest:
                if ext_map[extension_id][1]:
                    for version_range, ext_uris in ext_map[extension_id][1]:
                        if version_range.contains(stac_version):
                            if object_type in ext_uris:
                                uri = ext_uris[object_type]
                                break

        if uri is None:
            return uri
        else:
            return cls._append_base_uri_if_needed(uri, stac_version)



"""Implements the :stac-ext:`Electro-Optical Extension <eo>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions import projection, view
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.serialization.identify import STACJSONDescription, STACVersionID
from pystac.summaries import RangeSummary
from pystac.utils import get_required, map_opt

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset`,
#: pr :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI: str = "https://stac-extensions.github.io/eo/v1.1.0/schema.json"
SCHEMA_URIS: list[str] = [
    "https://stac-extensions.github.io/eo/v1.0.0/schema.json",
    SCHEMA_URI,
]
PREFIX: str = "eo:"

# Field names
BANDS_PROP: str = PREFIX + "bands"
CLOUD_COVER_PROP: str = PREFIX + "cloud_cover"
SNOW_COVER_PROP: str = PREFIX + "snow_cover"


def validated_percentage(v: float | None) -> float | None:
    if v is not None and not isinstance(v, (float, int)) or isinstance(v, bool):
        raise ValueError(f"Invalid percentage: {v} must be number")
    if v is not None and not 0 <= v <= 100:
        raise ValueError(f"Invalid percentage: {v} must be between 0 and 100")
    return v


class Band:
    """Represents Band information attached to an Item that implements the eo extension.

    Use :meth:`Band.create` to create a new Band.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(
        self,
        name: str,
        common_name: str | None = None,
        description: str | None = None,
        center_wavelength: float | None = None,
        full_width_half_max: float | None = None,
        solar_illumination: float | None = None,
    ) -> None:
        """
        Sets the properties for this Band.

        Args:
            name : The name of the band (e.g., "B01", "B02", "B1", "B5", "QA").
            common_name : The name commonly used to refer to the band to make it
                easier to search for bands across instruments. See the :stac-ext:`list
                of accepted common names <eo#common-band-names>`.
            description : Description to fully explain the band.
            center_wavelength : The center wavelength of the band, in micrometers (μm).
            full_width_half_max : Full width at half maximum (FWHM). The width of the
                band, as measured at half the maximum transmission, in micrometers (μm).
            solar_illumination: The solar illumination of the band,
                as measured at half the maximum transmission, in W/m2/micrometers.
        """
        self.name = name
        self.common_name = common_name
        self.description = description
        self.center_wavelength = center_wavelength
        self.full_width_half_max = full_width_half_max
        self.solar_illumination = solar_illumination

    @classmethod
    def create(
        cls,
        name: str,
        common_name: str | None = None,
        description: str | None = None,
        center_wavelength: float | None = None,
        full_width_half_max: float | None = None,
        solar_illumination: float | None = None,
    ) -> Band:
        """
        Creates a new band.

        Args:
            name : The name of the band (e.g., "B01", "B02", "B1", "B5", "QA").
            common_name : The name commonly used to refer to the band to make it easier
                to search for bands across instruments. See the :stac-ext:`list of
                accepted common names <eo#common-band-names>`.
            description : Description to fully explain the band.
            center_wavelength : The center wavelength of the band, in micrometers (μm).
            full_width_half_max : Full width at half maximum (FWHM). The width of the
                band, as measured at half the maximum transmission, in micrometers (μm).
            solar_illumination: The solar illumination of the band,
                as measured at half the maximum transmission, in W/m2/micrometers.
        """
        b = cls({})
        b.apply(
            name=name,
            common_name=common_name,
            description=description,
            center_wavelength=center_wavelength,
            full_width_half_max=full_width_half_max,
            solar_illumination=solar_illumination,
        )
        return b

    @property
    def name(self) -> str:
        """Get or sets the name of the band (e.g., "B01", "B02", "B1", "B5", "QA").

        Returns:
            str
        """
        return cast(str, get_required(self.properties.get("name"), self, "name"))

    @name.setter
    def name(self, v: str) -> None:
        self.properties["name"] = v

    @property
    def common_name(self) -> str | None:
        """Get or sets the name commonly used to refer to the band to make it easier
            to search for bands across instruments. See the :stac-ext:`list of accepted
            common names <eo#common-band-names>`.

        Returns:
            Optional[str]
        """
        return self.properties.get("common_name")

    @common_name.setter
    def common_name(self, v: str | None) -> None:
        if v is not None:
            self.properties["common_name"] = v
        else:
            self.properties.pop("common_name", None)

    @property
    def description(self) -> str | None:
        """Get or sets the description to fully explain the band. CommonMark 0.29
        syntax MAY be used for rich text representation.

        Returns:
            str
        """
        return self.properties.get("description")

    @description.setter
    def description(self, v: str | None) -> None:
        if v is not None:
            self.properties["description"] = v
        else:
            self.properties.pop("description", None)

    @property
    def center_wavelength(self) -> float | None:
        """Get or sets the center wavelength of the band, in micrometers (μm).

        Returns:
            float
        """
        return self.properties.get("center_wavelength")

    @center_wavelength.setter
    def center_wavelength(self, v: float | None) -> None:
        if v is not None:
            self.properties["center_wavelength"] = v
        else:
            self.properties.pop("center_wavelength", None)

    @property
    def full_width_half_max(self) -> float | None:
        """Get or sets the full width at half maximum (FWHM). The width of the band,
            as measured at half the maximum transmission, in micrometers (μm).

        Returns:
            [float]
        """
        return self.properties.get("full_width_half_max")

    @full_width_half_max.setter
    def full_width_half_max(self, v: float | None) -> None:
        if v is not None:
            self.properties["full_width_half_max"] = v
        else:
            self.properties.pop("full_width_half_max", None)

    @property
    def solar_illumination(self) -> float | None:
        """Get or sets the solar illumination of the band,
            as measured at half the maximum transmission, in W/m2/micrometers.

        Returns:
            [float]
        """
        return self.properties.get("solar_illumination")

    @solar_illumination.setter
    def solar_illumination(self, v: float | None) -> None:
        if v is not None:
            self.properties["solar_illumination"] = v
        else:
            self.properties.pop("solar_illumination", None)

    def __repr__(self) -> str:
        return f"<Band name={self.properties.get('name')}>"

    def to_dict(self) -> dict[str, Any]:
        """Returns this band as a dictionary.

        Returns:
            dict: The serialization of this Band.
        """
        return self.properties

    @staticmethod
    def band_range(common_name: str) -> tuple[float, float] | None:
        """Gets the band range for a common band name.

        Args:
            common_name : The common band name. Must be one of the :stac-ext:`list of
                accepted common names <eo#common-band-names>`.

        Returns:
            Tuple[float, float] or None: The band range for this name as (min, max), or
            None if this is not a recognized common name.
        """
        name_to_range = {
            "coastal": (0.40, 0.45),
            "blue": (0.45, 0.50),
            "green": (0.50, 0.60),
            "red": (0.60, 0.70),
            "yellow": (0.58, 0.62),
            "pan": (0.50, 0.70),
            "rededge": (0.70, 0.75),
            "nir": (0.75, 1.00),
            "nir08": (0.75, 0.90),
            "nir09": (0.85, 1.05),
            "cirrus": (1.35, 1.40),
            "swir16": (1.55, 1.75),
            "swir22": (2.10, 2.30),
            "lwir": (10.5, 12.5),
            "lwir11": (10.5, 11.5),
            "lwir12": (11.5, 12.5),
        }

        return name_to_range.get(common_name)

    @staticmethod
    def band_description(common_name: str) -> str | None:
        """Returns a description of the band for one with a common name.

        Args:
            common_name : The common band name. Must be one of the :stac-ext:`list of
                accepted common names <eo#common-band-names>`.

        Returns:
            str or None: If a recognized common name, returns a description including
            the band range. Otherwise, returns None.
        """
        r = Band.band_range(common_name)
        if r is not None:
            return f"Common name: {common_name}, Range: {r[0]} to {r[1]}"
        return None


class EOExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or :class:`~pystac.Asset` with properties from the
    :stac-ext:`Electro-Optical Extension <eo>`. This class is generic over the type of
    STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`EOExtension`, use the
    :meth:`EOExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> eo_ext = EOExtension.ext(item)
    """

    name: Literal["eo"] = "eo"

    def apply(
        self,
        bands: list[Band] | None = None,
        cloud_cover: float | None = None,
        snow_cover: float | None = None,
    ) -> None:
        """Applies Electro-Optical Extension properties to the extended
        :class:`~pystac.Item` or :class:`~pystac.Asset`.

        Args:
            bands : A list of available bands where each item is a :class:`~Band`
                object. If given, requires at least one band.
            cloud_cover : The estimate of cloud cover as a percentage
                (0-100) of the entire scene. If not available the field should not
                be provided.
            snow_cover : The estimate of snow cover as a percentage
                (0-100) of the entire scene. If not available the field should not
                be provided.
        """
        self.bands = bands
        self.cloud_cover = validated_percentage(cloud_cover)
        self.snow_cover = validated_percentage(snow_cover)

    @property
    def bands(self) -> list[Band] | None:
        """Gets or sets a list of available bands where each item is a :class:`~Band`
        object (or ``None`` if no bands have been set). If not available the field
        should not be provided.
        """
        return self._get_bands()

    @bands.setter
    def bands(self, v: list[Band] | None) -> None:
        self._set_property(
            BANDS_PROP, map_opt(lambda bands: [b.to_dict() for b in bands], v)
        )

    def _get_bands(self) -> list[Band] | None:
        return map_opt(
            lambda bands: [Band(b) for b in bands],
            self._get_property(BANDS_PROP, list[dict[str, Any]]),
        )

    @property
    def cloud_cover(self) -> float | None:
        """Get or sets the estimate of cloud cover as a percentage
        (0-100) of the entire scene. If not available the field should not be provided.

        Returns:
            float or None
        """
        return self._get_property(CLOUD_COVER_PROP, float)

    @cloud_cover.setter
    def cloud_cover(self, v: float | None) -> None:
        self._set_property(CLOUD_COVER_PROP, validated_percentage(v), pop_if_none=True)

    @property
    def snow_cover(self) -> float | None:
        """Get or sets the estimate of snow cover as a percentage
        (0-100) of the entire scene. If not available the field should not be provided.

        Returns:
            float or None
        """
        return self._get_property(SNOW_COVER_PROP, float)

    @snow_cover.setter
    def snow_cover(self, v: float | None) -> None:
        self._set_property(SNOW_COVER_PROP, validated_percentage(v), pop_if_none=True)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        import warnings

        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return SCHEMA_URIS

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> EOExtension[T]:
        """Extends the given STAC Object with properties from the
        :stac-ext:`Electro-Optical Extension <eo>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(EOExtension[T], ItemEOExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(EOExtension[T], AssetEOExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(EOExtension[T], ItemAssetsEOExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesEOExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesEOExtension(obj)


class ItemEOExtension(EOExtension[pystac.Item]):
    """A concrete implementation of :class:`EOExtension` on an :class:`~pystac.Item`
    that extends the properties of the Item to include properties defined in the
    :stac-ext:`Electro-Optical Extension <eo>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`EOExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def _get_bands(self) -> list[Band] | None:
        """Get or sets a list of :class:`~pystac.Band` objects that represent
        the available bands.
        """
        bands = self._get_property(BANDS_PROP, list[dict[str, Any]])

        # get assets with eo:bands even if not in item
        if bands is None:
            asset_bands: list[dict[str, Any]] = []
            for _, value in self.item.get_assets().items():
                if BANDS_PROP in value.extra_fields:
                    asset_bands.extend(
                        cast(list[dict[str, Any]], value.extra_fields.get(BANDS_PROP))
                    )
            if any(asset_bands):
                bands = asset_bands

        if bands is not None:
            return [Band(b) for b in bands]
        return None

    def get_assets(
        self,
        name: str | None = None,
        common_name: str | None = None,
    ) -> dict[str, pystac.Asset]:
        """Get the item's assets where eo:bands are defined.

        Args:
            name: If set, filter the assets such that only those with a
                matching ``eo:band.name`` are returned.
            common_name: If set, filter the assets such that only those with a matching
                ``eo:band.common_name`` are returned.

        Returns:
            Dict[str, Asset]: A dictionary of assets that match ``name``
                and/or ``common_name`` if set or else all of this item's assets were
                eo:bands are defined.
        """
        kwargs = {"name": name, "common_name": common_name}
        return {
            key: asset
            for key, asset in self.item.get_assets().items()
            if BANDS_PROP in asset.extra_fields
            and all(
                v is None or any(v == b.get(k) for b in asset.extra_fields[BANDS_PROP])
                for k, v in kwargs.items()
            )
        }

    def __repr__(self) -> str:
        return f"<ItemEOExtension Item id={self.item.id}>"


class AssetEOExtension(EOExtension[pystac.Asset]):
    """A concrete implementation of :class:`EOExtension` on an :class:`~pystac.Asset`
    that extends the Asset fields to include properties defined in the
    :stac-ext:`Electro-Optical Extension <eo>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`EOExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def _get_bands(self) -> list[Band] | None:
        if BANDS_PROP not in self.properties:
            return None
        return list(
            map(
                lambda band: Band(band),
                cast(list[dict[str, Any]], self.properties.get(BANDS_PROP)),
            )
        )

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetEOExtension Asset href={self.asset_href}>"


class ItemAssetsEOExtension(EOExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def _get_bands(self) -> list[Band] | None:
        if BANDS_PROP not in self.properties:
            return None
        return list(
            map(
                lambda band: Band(band),
                cast(list[dict[str, Any]], self.properties.get(BANDS_PROP)),
            )
        )

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesEOExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Electro-Optical Extension <eo>`.
    """

    @property
    def bands(self) -> list[Band] | None:
        """Get or sets the summary of :attr:`EOExtension.bands` values
        for this Collection.
        """

        return map_opt(
            lambda bands: [Band(b) for b in bands],
            self.summaries.get_list(BANDS_PROP),
        )

    @bands.setter
    def bands(self, v: list[Band] | None) -> None:
        self._set_summary(BANDS_PROP, map_opt(lambda x: [b.to_dict() for b in x], v))

    @property
    def cloud_cover(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`EOExtension.cloud_cover` values
        for this Collection.
        """
        return self.summaries.get_range(CLOUD_COVER_PROP)

    @cloud_cover.setter
    def cloud_cover(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(CLOUD_COVER_PROP, v)

    @property
    def snow_cover(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`EOExtension.snow_cover` values
        for this Collection.
        """
        return self.summaries.get_range(SNOW_COVER_PROP)

    @snow_cover.setter
    def snow_cover(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(SNOW_COVER_PROP, v)


class EOExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "eo",
        *[uri for uri in SCHEMA_URIS if uri != SCHEMA_URI],
    }
    stac_object_types = {pystac.STACObjectType.ITEM}

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        if version < "0.9":
            # Some eo fields became common_metadata
            if (
                "eo:platform" in obj["properties"]
                and "platform" not in obj["properties"]
            ):
                obj["properties"]["platform"] = obj["properties"]["eo:platform"]
                del obj["properties"]["eo:platform"]

            if (
                "eo:instrument" in obj["properties"]
                and "instruments" not in obj["properties"]
            ):
                obj["properties"]["instruments"] = [obj["properties"]["eo:instrument"]]
                del obj["properties"]["eo:instrument"]

            if (
                "eo:constellation" in obj["properties"]
                and "constellation" not in obj["properties"]
            ):
                obj["properties"]["constellation"] = obj["properties"][
                    "eo:constellation"
                ]
                del obj["properties"]["eo:constellation"]

            # Some eo fields became view extension fields
            eo_to_view_fields = [
                "off_nadir",
                "azimuth",
                "incidence_angle",
                "sun_azimuth",
                "sun_elevation",
            ]

            for field in eo_to_view_fields:
                if f"eo:{field}" in obj["properties"]:
                    if "stac_extensions" not in obj:
                        obj["stac_extensions"] = []
                    if view.SCHEMA_URI not in obj["stac_extensions"]:
                        obj["stac_extensions"].append(view.SCHEMA_URI)
                    if f"view:{field}" not in obj["properties"]:
                        obj["properties"][f"view:{field}"] = obj["properties"][
                            f"eo:{field}"
                        ]
                        del obj["properties"][f"eo:{field}"]

            # eo:epsg became proj:epsg in Projection Extension <2.0.0 and became
            # proj:code in Projection Extension 2.0.0
            eo_epsg = PREFIX + "epsg"
            proj_epsg = projection.PREFIX + "epsg"
            proj_code = projection.PREFIX + "code"
            if (
                eo_epsg in obj["properties"]
                and proj_epsg not in obj["properties"]
                and proj_code not in obj["properties"]
            ):
                obj["stac_extensions"] = obj.get("stac_extensions", [])
                if set(obj["stac_extensions"]).intersection(
                    projection.ProjectionExtensionHooks.pre_2
                ):
                    obj["properties"][proj_epsg] = obj["properties"].pop(eo_epsg)
                else:
                    obj["properties"][proj_code] = (
                        f"EPSG:{obj['properties'].pop(eo_epsg)}"
                    )
                    if not projection.ProjectionExtensionHooks().has_extension(obj):
                        obj["stac_extensions"].append(
                            projection.ProjectionExtension.get_schema_uri()
                        )

                if not any(prop.startswith(PREFIX) for prop in obj["properties"]):
                    obj["stac_extensions"].remove(EOExtension.get_schema_uri())

        if version < "1.0.0-beta.1" and info.object_type == pystac.STACObjectType.ITEM:
            # gsd moved from eo to common metadata
            if "eo:gsd" in obj["properties"]:
                obj["properties"]["gsd"] = obj["properties"]["eo:gsd"]
                del obj["properties"]["eo:gsd"]

            # The way bands were declared in assets changed.
            # In 1.0.0-beta.1 they are inlined into assets as
            # opposed to having indices back into a property-level array.
            if "eo:bands" in obj["properties"]:
                bands = obj["properties"]["eo:bands"]
                for asset in obj["assets"].values():
                    if "eo:bands" in asset:
                        new_bands: list[dict[str, Any]] = []
                        for band_index in asset["eo:bands"]:
                            new_bands.append(bands[band_index])
                        asset["eo:bands"] = new_bands

        super().migrate(obj, version, info)


EO_EXTENSION_HOOKS: ExtensionHooks = EOExtensionHooks()



"""Implements the :stac-ext:`Raster Extension <raster>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import StringEnum, get_opt, get_required, map_opt

#: Generalized version of :class:`~pystac.Asset` or
#: :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI = "https://stac-extensions.github.io/raster/v1.1.0/schema.json"
SCHEMA_URIS = [
    "https://stac-extensions.github.io/raster/v1.0.0/schema.json",
    SCHEMA_URI,
]
SCHEMA_STARTWITH = "https://stac-extensions.github.io/raster/"
BANDS_PROP = "raster:bands"


class Sampling(StringEnum):
    AREA = "area"
    POINT = "point"


class DataType(StringEnum):
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    FLOAT16 = "float16"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    CINT16 = "cint16"
    CINT32 = "cint32"
    CFLOAT32 = "cfloat32"
    CFLOAT64 = "cfloat64"
    OTHER = "other"


class NoDataStrings(StringEnum):
    INF = "inf"
    NINF = "-inf"
    NAN = "nan"


class Statistics:
    """Represents statistics information attached to a band in the raster extension.

    Use Statistics.create to create a new Statistics instance.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, float | None]) -> None:
        self.properties = properties

    def apply(
        self,
        minimum: float | None = None,
        maximum: float | None = None,
        mean: float | None = None,
        stddev: float | None = None,
        valid_percent: float | None = None,
    ) -> None:
        """
        Sets the properties for this raster Band.

        Args:
            minimum : Minimum value of all the pixels in the band.
            maximum : Maximum value of all the pixels in the band.
            mean : Mean value of all the pixels in the band.
            stddev : Standard Deviation value of all the pixels in the band.
            valid_percent : Percentage of valid (not nodata) pixel.
        """
        self.minimum = minimum
        self.maximum = maximum
        self.mean = mean
        self.stddev = stddev
        self.valid_percent = valid_percent

    @classmethod
    def create(
        cls,
        minimum: float | None = None,
        maximum: float | None = None,
        mean: float | None = None,
        stddev: float | None = None,
        valid_percent: float | None = None,
    ) -> Statistics:
        """
        Creates a new band.

        Args:
            minimum : Minimum value of all the pixels in the band.
            maximum : Maximum value of all the pixels in the band.
            mean : Mean value of all the pixels in the band.
            stddev : Standard Deviation value of all the pixels in the band.
            valid_percent : Percentage of valid (not nodata) pixel.
        """
        b = cls({})
        b.apply(
            minimum=minimum,
            maximum=maximum,
            mean=mean,
            stddev=stddev,
            valid_percent=valid_percent,
        )
        return b

    @property
    def minimum(self) -> float | None:
        """Get or sets the minimum pixel value

        Returns:
            Optional[float]
        """
        return self.properties.get("minimum")

    @minimum.setter
    def minimum(self, v: float | None) -> None:
        if v is not None:
            self.properties["minimum"] = v
        else:
            self.properties.pop("minimum", None)

    @property
    def maximum(self) -> float | None:
        """Get or sets the maximum pixel value

        Returns:
            Optional[float]
        """
        return self.properties.get("maximum")

    @maximum.setter
    def maximum(self, v: float | None) -> None:
        if v is not None:
            self.properties["maximum"] = v
        else:
            self.properties.pop("maximum", None)

    @property
    def mean(self) -> float | None:
        """Get or sets the mean pixel value

        Returns:
            Optional[float]
        """
        return self.properties.get("mean")

    @mean.setter
    def mean(self, v: float | None) -> None:
        if v is not None:
            self.properties["mean"] = v
        else:
            self.properties.pop("mean", None)

    @property
    def stddev(self) -> float | None:
        """Get or sets the standard deviation pixel value

        Returns:
            Optional[float]
        """
        return self.properties.get("stddev")

    @stddev.setter
    def stddev(self, v: float | None) -> None:
        if v is not None:
            self.properties["stddev"] = v
        else:
            self.properties.pop("stddev", None)

    @property
    def valid_percent(self) -> float | None:
        """Get or sets the Percentage of valid (not nodata) pixel

        Returns:
            Optional[float]
        """
        return self.properties.get("valid_percent")

    @valid_percent.setter
    def valid_percent(self, v: float | None) -> None:
        if v is not None:
            self.properties["valid_percent"] = v
        else:
            self.properties.pop("valid_percent", None)

    def to_dict(self) -> dict[str, Any]:
        """Returns these statistics as a dictionary.

        Returns:
            dict: The serialization of the Statistics.
        """
        return self.properties

    @staticmethod
    def from_dict(d: dict[str, Any]) -> Statistics:
        """Constructs a Statistics from a dict.

        Returns:
            Statistics: The Statistics deserialized from the JSON dict.
        """
        return Statistics(properties=d)


class Histogram:
    """Represents pixel distribution information attached to a band in the raster
    extension.

    Use Band.create to create a new Band.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(
        self,
        count: int,
        min: float,
        max: float,
        buckets: list[int],
    ) -> None:
        """
        Sets the properties for this raster Band.

        Args:
            count : number of buckets of the distribution.
            min : minimum value of the distribution.
                Also, the mean value of the first bucket.
            max : maximum value of the distribution.
                Also, the mean value of the last bucket.
            buckets : Array of integer indicating the number
                of pixels included in the bucket.
        """
        self.count = count
        self.min = min
        self.max = max
        self.buckets = buckets

    @classmethod
    def create(
        cls,
        count: int,
        min: float,
        max: float,
        buckets: list[int],
    ) -> Histogram:
        """
        Creates a new band.

        Args:
            count : number of buckets of the distribution.
            min : minimum value of the distribution.
                Also, the mean value of the first bucket.
            max : maximum value of the distribution.
                Also, the mean value of the last bucket.
            buckets : Array of integer indicating the number
                of pixels included in the bucket.
        """
        b = cls({})
        b.apply(
            count=count,
            min=min,
            max=max,
            buckets=buckets,
        )
        return b

    @property
    def count(self) -> int:
        """Get or sets the number of buckets of the distribution.

        Returns:
            int
        """
        return cast(int, get_required(self.properties.get("count"), self, "count"))

    @count.setter
    def count(self, v: int) -> None:
        self.properties["count"] = v

    @property
    def min(self) -> float:
        """Get or sets the minimum value of the distribution.

        Returns:
            float
        """
        return cast(float, get_required(self.properties.get("min"), self, "min"))

    @min.setter
    def min(self, v: float) -> None:
        self.properties["min"] = v

    @property
    def max(self) -> float:
        """Get or sets the maximum value of the distribution.

        Returns:
            float
        """
        return cast(float, get_required(self.properties.get("max"), self, "max"))

    @max.setter
    def max(self, v: float) -> None:
        self.properties["max"] = v

    @property
    def buckets(self) -> list[int]:
        """Get or sets the Array of integer indicating
        the number of pixels included in the bucket.

        Returns:
            List[int]
        """
        return get_required(self.properties.get("buckets"), self, "buckets")

    @buckets.setter
    def buckets(self, v: list[int]) -> None:
        self.properties["buckets"] = v

    def to_dict(self) -> dict[str, Any]:
        """Returns this histogram as a dictionary.

        Returns:
            dict: The serialization of the Histogram.
        """
        return self.properties

    @staticmethod
    def from_dict(d: dict[str, Any]) -> Histogram:
        """Constructs a Histogram from a dict.

        Returns:
            Histogram: The Histogram deserialized from the JSON dict.
        """
        return Histogram(properties=d)


class RasterBand:
    """Represents a Raster Band information attached to an Item
    that implements the raster extension.

    Use Band.create to create a new Band.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(
        self,
        nodata: float | NoDataStrings | None = None,
        sampling: Sampling | None = None,
        data_type: DataType | None = None,
        bits_per_sample: float | None = None,
        spatial_resolution: float | None = None,
        statistics: Statistics | None = None,
        unit: str | None = None,
        scale: float | None = None,
        offset: float | None = None,
        histogram: Histogram | None = None,
    ) -> None:
        """
        Sets the properties for this raster Band.

        Args:
            nodata : Pixel values used to identify pixels that are nodata in the assets.
            sampling : One of area or point. Indicates whether a pixel value should be
                assumed to represent a sampling over the region of the pixel or a point
                sample at the center of the pixel.
            data_type : The data type of the band.
                One of the data types as described in the
                :stac-ext:`Raster Data Types <raster/#data-types> docs`.
            bits_per_sample : The actual number of bits used for this band.
                Normally only present when the number of bits is non-standard for the
                datatype, such as when a 1 bit TIFF is represented as byte
            spatial_resolution : Average spatial resolution (in meters) of the pixels in
                the band.
            statistics: Statistics of all the pixels in the band
            unit: unit denomination of the pixel value
            scale: multiplicator factor of the pixel value to transform into the value
                (i.e. translate digital number to reflectance).
            offset: number to be added to the pixel value (after scaling) to transform
                into the value (i.e. translate digital number to reflectance).
            histogram: Histogram distribution information of the pixels values in the
                band
        """
        self.nodata = nodata
        self.sampling = sampling
        self.data_type = data_type
        self.bits_per_sample = bits_per_sample
        self.spatial_resolution = spatial_resolution
        self.statistics = statistics
        self.unit = unit
        self.scale = scale
        self.offset = offset
        self.histogram = histogram

    @classmethod
    def create(
        cls,
        nodata: float | NoDataStrings | None = None,
        sampling: Sampling | None = None,
        data_type: DataType | None = None,
        bits_per_sample: float | None = None,
        spatial_resolution: float | None = None,
        statistics: Statistics | None = None,
        unit: str | None = None,
        scale: float | None = None,
        offset: float | None = None,
        histogram: Histogram | None = None,
    ) -> RasterBand:
        """
        Creates a new band.

        Args:
            nodata : Pixel values used to identify pixels that are nodata in the assets.
            sampling : One of area or point. Indicates whether a pixel value should be
                assumed to represent a sampling over the region of the pixel or a point
                sample at the center of the pixel.
            data_type :The data type of the band.
                One of the data types as described in the
                :stac-ext:`Raster Data Types <raster/#data-types> docs`.
            bits_per_sample : The actual number of bits used for this band.
                Normally only present when the number of bits is non-standard for the
                datatype, such as when a 1 bit TIFF is represented as byte
            spatial_resolution : Average spatial resolution (in meters) of the pixels in
                the band.
            statistics: Statistics of all the pixels in the band
            unit: unit denomination of the pixel value
            scale: multiplicator factor of the pixel value to transform into the value
                (i.e. translate digital number to reflectance).
            offset: number to be added to the pixel value (after scaling) to transform
                into the value (i.e. translate digital number to reflectance).
            histogram: Histogram distribution information of the pixels values in the
                band
        """
        b = cls({})
        b.apply(
            nodata=nodata,
            sampling=sampling,
            data_type=data_type,
            bits_per_sample=bits_per_sample,
            spatial_resolution=spatial_resolution,
            statistics=statistics,
            unit=unit,
            scale=scale,
            offset=offset,
            histogram=histogram,
        )
        return b

    @property
    def nodata(self) -> float | NoDataStrings | None:
        """Get or sets the nodata pixel value

        Returns:
            Optional[float]
        """
        return self.properties.get("nodata")

    @nodata.setter
    def nodata(self, v: float | NoDataStrings | None) -> None:
        if v is not None:
            self.properties["nodata"] = v
        else:
            self.properties.pop("nodata", None)

    @property
    def sampling(self) -> Sampling | None:
        """Get or sets the property indicating whether a pixel value should be assumed
        to represent a sampling over the region of the pixel or a point sample
        at the center of the pixel.

        Returns:
            Optional[Sampling]
        """
        return self.properties.get("sampling")

    @sampling.setter
    def sampling(self, v: Sampling | None) -> None:
        if v is not None:
            self.properties["sampling"] = v
        else:
            self.properties.pop("sampling", None)

    @property
    def data_type(self) -> DataType | None:
        """Get or sets the data type of the band.

        Returns:
            Optional[DataType]
        """
        return self.properties.get("data_type")

    @data_type.setter
    def data_type(self, v: DataType | None) -> None:
        if v is not None:
            self.properties["data_type"] = v
        else:
            self.properties.pop("data_type", None)

    @property
    def bits_per_sample(self) -> float | None:
        """Get or sets the actual number of bits used for this band.

        Returns:
            float
        """
        return self.properties.get("bits_per_sample")

    @bits_per_sample.setter
    def bits_per_sample(self, v: float | None) -> None:
        if v is not None:
            self.properties["bits_per_sample"] = v
        else:
            self.properties.pop("bits_per_sample", None)

    @property
    def spatial_resolution(self) -> float | None:
        """Get or sets the average spatial resolution (in meters) of the pixels in the
        band.

        Returns:
            [float]
        """
        return self.properties.get("spatial_resolution")

    @spatial_resolution.setter
    def spatial_resolution(self, v: float | None) -> None:
        if v is not None:
            self.properties["spatial_resolution"] = v
        else:
            self.properties.pop("spatial_resolution", None)

    @property
    def statistics(self) -> Statistics | None:
        """Get or sets the average spatial resolution (in meters) of the pixels in the
        band.

        Returns:
            [Statistics]
        """
        return Statistics.from_dict(get_opt(self.properties.get("statistics")))

    @statistics.setter
    def statistics(self, v: Statistics | None) -> None:
        if v is not None:
            self.properties["statistics"] = v.to_dict()
        else:
            self.properties.pop("statistics", None)

    @property
    def unit(self) -> str | None:
        """Get or sets the unit denomination of the pixel value

        Returns:
            [str]
        """
        return self.properties.get("unit")

    @unit.setter
    def unit(self, v: str | None) -> None:
        if v is not None:
            self.properties["unit"] = v
        else:
            self.properties.pop("unit", None)

    @property
    def scale(self) -> float | None:
        """Get or sets the multiplicator factor of the pixel value to transform
        into the value (i.e. translate digital number to reflectance).

        Returns:
            [float]
        """
        return self.properties.get("scale")

    @scale.setter
    def scale(self, v: float | None) -> None:
        if v is not None:
            self.properties["scale"] = v
        else:
            self.properties.pop("scale", None)

    @property
    def offset(self) -> float | None:
        """Get or sets the number to be added to the pixel value (after scaling)
        to transform into the value (i.e. translate digital number to reflectance).

        Returns:
            [float]
        """
        return self.properties.get("offset")

    @offset.setter
    def offset(self, v: float | None) -> None:
        if v is not None:
            self.properties["offset"] = v
        else:
            self.properties.pop("offset", None)

    @property
    def histogram(self) -> Histogram | None:
        """Get or sets the histogram distribution information of the pixels values in
        the band.

        Returns:
            [Histogram]
        """
        return Histogram.from_dict(get_opt(self.properties.get("histogram")))

    @histogram.setter
    def histogram(self, v: Histogram | None) -> None:
        if v is not None:
            self.properties["histogram"] = v.to_dict()
        else:
            self.properties.pop("histogram", None)

    def __repr__(self) -> str:
        return "<Raster Band>"

    def to_dict(self) -> dict[str, Any]:
        """Returns this band as a dictionary.

        Returns:
            dict: The serialization of the Band.
        """
        return self.properties


class RasterExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Asset`, or
    :class:`~pystac.ItemAssetDefinition` with properties from
    the :stac-ext:`Raster Extension <raster>`. This class is generic over
    the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    This class will generally not be used directly. Instead, use the concrete
    implementation associated with the STAC Object you want to extend (e.g.
    :class:`~AssetRasterExtension` to extend an :class:`~pystac.Item`).  You may
    prefer to use the `ext` class method of this class to construct the correct
    instance type for you.
    """

    name: Literal["raster"] = "raster"

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    def apply(self, bands: list[RasterBand]) -> None:
        """Applies raster extension properties to the extended :class:`pystac.Item` or
        :class:`pystac.Asset`.

        Args:
            bands : a list of :class:`RasterBand` objects that represent
                the available raster bands.
        """
        self.bands = bands

    @property
    def bands(self) -> list[RasterBand] | None:
        """Gets or sets a list of available bands where each item is a
        :class:`~RasterBand` object (or ``None`` if no bands have been set). If not
        available the field should not be provided.
        """
        return self._get_bands()

    @bands.setter
    def bands(self, v: list[RasterBand] | None) -> None:
        self._set_property(
            BANDS_PROP, map_opt(lambda bands: [b.to_dict() for b in bands], v)
        )

    def _get_bands(self) -> list[RasterBand] | None:
        return map_opt(
            lambda bands: [RasterBand(b) for b in bands],
            self._get_property(BANDS_PROP, list[dict[str, Any]]),
        )

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        import warnings

        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return SCHEMA_URIS

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> RasterExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Raster
        Extension <raster>`.

        This extension can be applied to instances of :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(RasterExtension[T], AssetRasterExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(RasterExtension[T], ItemAssetsRasterExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesRasterExtension:
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesRasterExtension(obj)


class AssetRasterExtension(RasterExtension[pystac.Asset]):
    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetRasterExtension Asset href={self.asset_href}>"


class ItemAssetsRasterExtension(RasterExtension[pystac.ItemAssetDefinition]):
    asset_definition: pystac.ItemAssetDefinition
    """A reference to the :class:`~pystac.ItemAssetDefinition`
    being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.ItemAssetDefinition` fields, including
    extension properties."""

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.properties = item_asset.properties
        self.asset_definition = item_asset

    def __repr__(self) -> str:
        return "<ItemAssetsRasterExtension AssetDefinition={}>".format(
            self.asset_definition
        )


class SummariesRasterExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Raster Extension <raster>`.
    """

    @property
    def bands(self) -> list[RasterBand] | None:
        """Get or sets a list of :class:`RasterBand` objects that represent
        the available bands.
        """
        return map_opt(
            lambda bands: [RasterBand(b) for b in bands],
            self.summaries.get_list(BANDS_PROP),
        )

    @bands.setter
    def bands(self, v: list[RasterBand] | None) -> None:
        self._set_summary(BANDS_PROP, map_opt(lambda x: [b.to_dict() for b in x], v))


class RasterExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: set[str] = {*[uri for uri in SCHEMA_URIS if uri != SCHEMA_URI]}
    stac_object_types = {pystac.STACObjectType.ITEM, pystac.STACObjectType.COLLECTION}


RASTER_EXTENSION_HOOKS: ExtensionHooks = RasterExtensionHooks()



import os
from pathlib import Path

import pytest

import pystac


@pytest.mark.parametrize("action", ["copy", "move"])
def test_alter_asset_absolute_path(
    action: str, tmp_asset: pystac.Asset, tmp_path: Path
) -> None:
    asset = tmp_asset
    old_href = asset.get_absolute_href()
    assert old_href is not None

    new_href = str(tmp_path / "data.geojson")
    getattr(asset, action)(new_href)

    assert asset.href == new_href
    assert asset.get_absolute_href() == new_href
    assert os.path.exists(new_href)
    if action == "move":
        assert not os.path.exists(old_href)
    elif action == "copy":
        assert os.path.exists(old_href)


@pytest.mark.parametrize("action", ["copy", "move"])
def test_alter_asset_relative_path(action: str, tmp_asset: pystac.Asset) -> None:
    asset = tmp_asset
    old_href = asset.get_absolute_href()
    assert old_href is not None

    new_href = "./different.geojson"
    getattr(asset, action)(new_href)

    assert asset.href == new_href
    href = asset.get_absolute_href()
    assert href is not None
    assert os.path.exists(href)
    if action == "move":
        assert not os.path.exists(old_href)
    elif action == "copy":
        assert os.path.exists(old_href)


@pytest.mark.parametrize("action", ["copy", "move"])
def test_alter_asset_relative_src_no_owner_fails(
    action: str, tmp_asset: pystac.Asset
) -> None:
    asset = tmp_asset
    asset.owner = None
    new_href = "./different.geojson"
    with pytest.raises(ValueError, match=f"Cannot {action} file") as e:
        getattr(asset, action)(new_href)

    assert new_href not in str(e.value)
    assert asset.href != new_href


@pytest.mark.parametrize("action", ["copy", "move"])
def test_alter_asset_relative_dst_no_owner_fails(
    action: str, tmp_asset: pystac.Asset
) -> None:
    asset = tmp_asset
    item = asset.owner

    assert isinstance(item, pystac.Item)
    item.make_asset_hrefs_absolute()

    asset.owner = None
    new_href = "./different.geojson"
    with pytest.raises(ValueError, match=f"Cannot {action} file") as e:
        getattr(asset, action)(new_href)

    assert new_href in str(e.value)
    assert asset.href != new_href


def test_delete_asset(tmp_asset: pystac.Asset) -> None:
    asset = tmp_asset
    href = asset.get_absolute_href()
    assert href is not None
    assert os.path.exists(href)

    asset.delete()

    assert not os.path.exists(href)


def test_delete_asset_relative_no_owner_fails(tmp_asset: pystac.Asset) -> None:
    asset = tmp_asset
    href = asset.get_absolute_href()
    assert href is not None
    assert os.path.exists(href)

    asset.owner = None

    with pytest.raises(ValueError, match="Cannot delete file") as e:
        asset.delete()

    assert asset.href in str(e.value)
    assert os.path.exists(href)


@pytest.mark.parametrize(
    "self_href, asset_href, expected_href",
    (
        (
            "http://test.com/stac/catalog/myitem.json",
            "asset.data",
            "http://test.com/stac/catalog/asset.data",
        ),
        (
            "http://test.com/stac/catalog/myitem.json",
            "/asset.data",
            "http://test.com/asset.data",
        ),
    ),
)
def test_asset_get_absolute_href(
    tmp_asset: pystac.Asset,
    self_href: str,
    asset_href: str,
    expected_href: str,
) -> None:
    asset = tmp_asset
    item = asset.owner

    if not isinstance(item, pystac.Item):
        raise TypeError("Asset must belong to an Item")

    # Set the item HREF as per test
    item.set_self_href(self_href)
    assert item.get_self_href() == self_href

    # Set the asset HREF as per test and check expected output
    asset.href = asset_href
    assert asset.get_absolute_href() == expected_href


@pytest.mark.skipif(os.name == "nt", reason="Unix only test")
@pytest.mark.parametrize(
    "self_href, asset_href, expected_href",
    (
        (
            "/local/myitem.json",
            "asset.data",
            "/local/asset.data",
        ),
        (
            "/local/myitem.json",
            "subdir/asset.data",
            "/local/subdir/asset.data",
        ),
        (
            "/local/myitem.json",
            "/absolute/asset.data",
            "/absolute/asset.data",
        ),
    ),
)
def test_asset_get_absolute_href_unix(
    tmp_asset: pystac.Asset,
    self_href: str,
    asset_href: str,
    expected_href: str,
) -> None:
    test_asset_get_absolute_href(tmp_asset, self_href, asset_href, expected_href)


@pytest.mark.skipif(os.name != "nt", reason="Windows only test")
@pytest.mark.parametrize(
    "self_href, asset_href, expected_href",
    (
        (
            "{tmpdir}/myitem.json",
            "asset.data",
            "{tmpdir}/asset.data",
        ),
        (
            "{tmpdir}/myitem.json",
            "subdir/asset.data",
            "{tmpdir}/subdir/asset.data",
        ),
        (
            "{tmpdir}/myitem.json",
            "c:/absolute/asset.data",
            "c:/absolute/asset.data",
        ),
        (
            "{tmpdir}/myitem.json",
            "d:\\absolute\\asset.data",
            "d:\\absolute\\asset.data",
        ),
    ),
)
def test_asset_get_absolute_href_windows(
    tmp_path: Path,
    tmp_asset: pystac.Asset,
    self_href: str,
    asset_href: str,
    expected_href: str,
) -> None:
    # For windows, we need an actual existing temporary directory
    tmpdir = tmp_path.as_posix()
    test_asset_get_absolute_href(
        tmp_asset,
        self_href.format(tmpdir=tmpdir),
        asset_href.format(tmpdir=tmpdir),
        expected_href.format(tmpdir=tmpdir),
    )



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast

import pystac
from pystac import utils
from pystac.errors import STACError

if TYPE_CHECKING:
    from pystac.asset import Asset
    from pystac.item import Item
    from pystac.provider import Provider


P = TypeVar("P")


class CommonMetadata:
    """Object containing fields that are not included in core item schema but
    are still commonly used. All attributes are defined within the properties of
    this item and are optional

    Args:
        properties : Dictionary of attributes that is the Item's properties
    """

    object: Asset | Item
    """The object from which common metadata is obtained."""

    def __init__(self, object: Asset | Item):
        self.object = object

    def _set_field(self, prop_name: str, v: Any | None) -> None:
        if hasattr(self.object, prop_name):
            setattr(self.object, prop_name, v)
        elif hasattr(self.object, "properties"):
            item = cast(pystac.Item, self.object)
            if v is None:
                item.properties.pop(prop_name, None)
            else:
                item.properties[prop_name] = v
        elif hasattr(self.object, "extra_fields") and isinstance(
            self.object.extra_fields, dict
        ):
            if v is None:
                self.object.extra_fields.pop(prop_name, None)
            else:
                self.object.extra_fields[prop_name] = v
        else:
            raise pystac.STACError(f"Cannot set field {prop_name} on {self}.")

    def _get_field(self, prop_name: str, _typ: type[P]) -> P | None:
        if hasattr(self.object, prop_name):
            return cast(Optional[P], getattr(self.object, prop_name))
        elif hasattr(self.object, "properties"):
            item = cast(pystac.Item, self.object)
            return item.properties.get(prop_name)
        elif hasattr(self.object, "extra_fields") and isinstance(
            self.object.extra_fields, dict
        ):
            return self.object.extra_fields.get(prop_name)
        else:
            raise STACError(f"Cannot get field {prop_name} from {self}.")

    # Basics
    @property
    def title(self) -> str | None:
        """Gets or set the object's title."""
        return self._get_field("title", str)

    @title.setter
    def title(self, v: str | None) -> None:
        self._set_field("title", v)

    @property
    def description(self) -> str | None:
        """Gets or set the object's description."""
        return self._get_field("description", str)

    @description.setter
    def description(self, v: str | None) -> None:
        if v == "":
            raise ValueError("description cannot be an empty string")
        self._set_field("description", v)

    # Date and Time Range
    @property
    def start_datetime(self) -> datetime | None:
        """Get or set the object's start_datetime.

        Note:
            ``start_datetime`` is an inclusive datetime.
        """
        return utils.map_opt(
            utils.str_to_datetime, self._get_field("start_datetime", str)
        )

    @start_datetime.setter
    def start_datetime(self, v: datetime | None) -> None:
        self._set_field("start_datetime", utils.map_opt(utils.datetime_to_str, v))

    @property
    def end_datetime(self) -> datetime | None:
        """Get or set the item's end_datetime.

        Note:
            ``end_datetime`` is an inclusive datetime.
        """
        return utils.map_opt(
            utils.str_to_datetime, self._get_field("end_datetime", str)
        )

    @end_datetime.setter
    def end_datetime(self, v: datetime | None) -> None:
        self._set_field("end_datetime", utils.map_opt(utils.datetime_to_str, v))

    # License
    @property
    def license(self) -> str | None:
        """Get or set the current license. License should be provided
        as a `SPDX License identifier <https://spdx.org/licenses/>`_,
        or `other`. If object includes data with multiple
        different licenses, use `other` and add a link for
        each.

        Note:
            The licenses `various` and `proprietary` are deprecated.
        """
        return self._get_field("license", str)

    @license.setter
    def license(self, v: str | None) -> None:
        self._set_field("license", v)

    # Providers
    @property
    def providers(self) -> list[Provider] | None:
        """Get or set a list of the object's providers."""
        return utils.map_opt(
            lambda providers: [pystac.Provider.from_dict(d) for d in providers],
            self._get_field("providers", list[dict[str, Any]]),
        )

    @providers.setter
    def providers(self, v: list[Provider] | None) -> None:
        self._set_field(
            "providers",
            utils.map_opt(lambda providers: [p.to_dict() for p in providers], v),
        )

    # Instrument
    @property
    def platform(self) -> str | None:
        """Gets or set the object's platform attribute."""
        return self._get_field("platform", str)

    @platform.setter
    def platform(self, v: str | None) -> None:
        self._set_field("platform", v)

    @property
    def instruments(self) -> list[str] | None:
        """Gets or sets the names of the instruments used."""
        return self._get_field("instruments", list[str])

    @instruments.setter
    def instruments(self, v: list[str] | None) -> None:
        self._set_field("instruments", v)

    @property
    def constellation(self) -> str | None:
        """Gets or set the name of the constellation associate with an object."""
        return self._get_field("constellation", str)

    @constellation.setter
    def constellation(self, v: str | None) -> None:
        self._set_field("constellation", v)

    @property
    def mission(self) -> str | None:
        """Gets or set the name of the mission associated with an object."""
        return self._get_field("mission", str)

    @mission.setter
    def mission(self, v: str | None) -> None:
        self._set_field("mission", v)

    @property
    def gsd(self) -> float | None:
        """Gets or sets the Ground Sample Distance at the sensor."""
        return self._get_field("gsd", float)

    @gsd.setter
    def gsd(self, v: float | None) -> None:
        self._set_field("gsd", v)

    # Metadata
    @property
    def created(self) -> datetime | None:
        """Get or set the metadata file's creation date. All datetime attributes have
        setters that can take either a string or a datetime, but always stores
        the attribute as a string.

        Note:
            ``created`` has a different meaning depending on the type of STAC object.
            On an :class:`~pystac.Item`, it refers to the creation time of the
            metadata. On an :class:`~pystac.Asset`, it refers to the creation time of
            the actual data linked to in :attr:`~pystac.Asset.href`.
        """
        return utils.map_opt(utils.str_to_datetime, self._get_field("created", str))

    @created.setter
    def created(self, v: datetime | None) -> None:
        self._set_field("created", utils.map_opt(utils.datetime_to_str, v))

    @property
    def updated(self) -> datetime | None:
        """Get or set the metadata file's update date. All datetime attributes have
        setters that can take either a string or a datetime, but always stores
        the attribute as a string

        Note:
            ``updated`` has a different meaning depending on the type of STAC object.
            On an :class:`~pystac.Item`, it refers to the update time of the
            metadata. On an :class:`~pystac.Asset`, it refers to the update time of
            the actual data linked to in :attr:`~pystac.Asset.href`.
        """
        return utils.map_opt(utils.str_to_datetime, self._get_field("updated", str))

    @updated.setter
    def updated(self, v: datetime | None) -> None:
        self._set_field("updated", utils.map_opt(utils.datetime_to_str, v))

    @property
    def keywords(self) -> list[str] | None:
        """Get or set the keywords describing the STAC entity."""
        return self._get_field("keywords", list[str])

    @keywords.setter
    def keywords(self, v: list[str] | None) -> None:
        self._set_field("keywords", v)

    @property
    def roles(self) -> list[str] | None:
        """Get or set the semantic roles of the entity."""
        return self._get_field("roles", list[str])

    @roles.setter
    def roles(self, v: list[str] | None) -> None:
        self._set_field("roles", v)



import json
from copy import deepcopy
from os.path import relpath
from typing import Any, cast

import pytest

import pystac
from pystac import Item, StacIO
from pystac.item_collection import ItemCollection
from tests.utils import TestCases
from tests.utils.stac_io_mock import MockDefaultStacIO

SIMPLE_ITEM = TestCases.get_path("data-files/examples/1.0.0-RC1/simple-item.json")
CORE_ITEM = TestCases.get_path("data-files/examples/1.0.0-RC1/core-item.json")
EXTENDED_ITEM = TestCases.get_path("data-files/examples/1.0.0-RC1/extended-item.json")

ITEM_COLLECTION = TestCases.get_path(
    "data-files/item-collection/sample-item-collection.json"
)


@pytest.fixture
def item_collection_dict() -> dict[str, Any]:
    with open(ITEM_COLLECTION) as src:
        return cast(dict[str, Any], json.load(src))


@pytest.fixture
def items(item_collection_dict: dict[str, Any]) -> list[Item]:
    return [Item.from_dict(f) for f in item_collection_dict["features"]]


@pytest.fixture
def stac_io() -> StacIO:
    return StacIO.default()


def test_item_collection_length(
    item_collection_dict: dict[str, Any], items: list[Item]
) -> None:
    item_collection = pystac.ItemCollection(items=items)

    assert len(item_collection) == len(items)


def test_item_collection_iter(items: list[Item]) -> None:
    expected_ids = [item.id for item in items]
    item_collection = pystac.ItemCollection(items=items)

    actual_ids = [item.id for item in item_collection]

    assert expected_ids == actual_ids


def test_item_collection_get_item_by_index(items: list[Item]) -> None:
    expected_id = items[0].id
    item_collection = pystac.ItemCollection(items=items)

    assert item_collection[0].id == expected_id


def test_item_collection_contains() -> None:
    item = pystac.Item.from_file(SIMPLE_ITEM)
    item_collection = pystac.ItemCollection(items=[item], clone_items=False)

    assert item in item_collection


def test_item_collection_extra_fields(items: list[Item]) -> None:
    item_collection = pystac.ItemCollection(
        items=items, extra_fields={"custom_field": "My value"}
    )

    assert item_collection.extra_fields.get("custom_field") == "My value"


def test_item_collection_to_dict(items: list[Item]) -> None:
    item_collection = pystac.ItemCollection(
        items=items, extra_fields={"custom_field": "My value"}
    )

    d = item_collection.to_dict()

    assert len(d["features"]) == len(items)
    assert d.get("custom_field") == "My value"


def test_item_collection_from_dict(items: list[Item]) -> None:
    features = [item.to_dict(transform_hrefs=False) for item in items]
    d = {
        "type": "FeatureCollection",
        "features": features,
        "custom_field": "My value",
    }
    item_collection = pystac.ItemCollection.from_dict(d)
    expected = len(features)
    assert expected == len(item_collection.items)
    assert item_collection.extra_fields.get("custom_field") == "My value"


def test_clone_item_collection() -> None:
    item_collection_1 = pystac.ItemCollection.from_file(ITEM_COLLECTION)
    item_collection_2 = item_collection_1.clone()

    item_ids_1 = [item.id for item in item_collection_1]
    item_ids_2 = [item.id for item in item_collection_2]

    # All items from the original collection should be in the clone...
    assert item_ids_1 == item_ids_2
    # ... but they should not be the same objects
    assert item_collection_1[0] is not item_collection_2[0]


def test_raise_error_for_invalid_object(stac_io: StacIO) -> None:
    item_dict = stac_io.read_json(SIMPLE_ITEM)

    with pytest.raises(pystac.STACTypeError):
        _ = pystac.ItemCollection.from_dict(item_dict)


def test_from_relative_path() -> None:
    _ = pystac.ItemCollection.from_file(
        relpath(
            TestCases.get_path("data-files/item-collection/sample-item-collection.json")
        )
    )


def test_from_list_of_dicts(stac_io: StacIO) -> None:
    item_dict = stac_io.read_json(SIMPLE_ITEM)
    item_collection = pystac.ItemCollection(items=[item_dict], clone_items=True)

    assert item_collection[0].id == item_dict.get("id")


def test_add_item_collections() -> None:
    item_1 = pystac.Item.from_file(SIMPLE_ITEM)
    item_2 = pystac.Item.from_file(EXTENDED_ITEM)
    item_3 = pystac.Item.from_file(CORE_ITEM)

    item_collection_1 = pystac.ItemCollection(items=[item_1, item_2])
    item_collection_2 = pystac.ItemCollection(items=[item_2, item_3])

    combined = item_collection_1 + item_collection_2

    assert len(combined) == 4


def test_add_other_raises_error() -> None:
    item_collection = pystac.ItemCollection.from_file(ITEM_COLLECTION)

    with pytest.raises(TypeError):
        _ = item_collection + 2


def test_identify_0_8_itemcollection_type(stac_io: StacIO) -> None:
    itemcollection_path = TestCases.get_path(
        "data-files/examples/0.8.1/item-spec/examples/itemcollection-sample-full.json"
    )
    itemcollection_dict = stac_io.read_json(itemcollection_path)

    assert pystac.ItemCollection.is_item_collection(itemcollection_dict), (
        "Did not correctly identify valid STAC 0.8 ItemCollection."
    )


def test_identify_0_9_itemcollection(stac_io: StacIO) -> None:
    itemcollection_path = TestCases.get_path(
        "data-files/examples/0.9.0/item-spec/examples/itemcollection-sample-full.json"
    )
    itemcollection_dict = stac_io.read_json(itemcollection_path)

    assert pystac.ItemCollection.is_item_collection(itemcollection_dict), (
        "Did not correctly identify valid STAC 0.9 ItemCollection."
    )


def test_from_dict_preserves_dict(item_collection_dict: dict[str, Any]) -> None:
    param_dict = deepcopy(item_collection_dict)

    # test that the parameter is preserved
    _ = ItemCollection.from_dict(param_dict)
    assert param_dict == item_collection_dict

    # assert that the parameter is preserved regardless of
    # preserve_dict
    _ = ItemCollection.from_dict(param_dict, preserve_dict=False)
    assert param_dict == item_collection_dict


def test_from_dict_sets_root(item_collection_dict: dict[str, Any]) -> None:
    param_dict = deepcopy(item_collection_dict)
    catalog = pystac.Catalog(id="test", description="test desc")
    item_collection = ItemCollection.from_dict(param_dict, root=catalog)
    for item in item_collection.items:
        assert item.get_root() == catalog


def test_to_dict_does_not_read_root_link_of_items() -> None:
    with MockDefaultStacIO() as mock_stac_io:
        item_collection = pystac.ItemCollection.from_file(ITEM_COLLECTION)

        item_collection.to_dict()

        assert mock_stac_io.mock.read_text.call_count == 1



import os
from collections.abc import Generator
from unittest.mock import patch

import pytest

import pystac
from tests.utils import TestCases


def test_override_stac_version_with_environ() -> None:
    override_version = "1.0.0-gamma.2"
    with patch.dict(os.environ, {"PYSTAC_STAC_VERSION_OVERRIDE": override_version}):
        cat = TestCases.case_1()
        d = cat.to_dict()
    assert d["stac_version"] == override_version


@pytest.fixture
def override_pystac_version() -> Generator[str]:
    stac_version = pystac.get_stac_version()
    override_version = "1.0.0-delta.2"
    pystac.set_stac_version(override_version)
    yield override_version
    pystac.set_stac_version(stac_version)


def test_override_stac_version_with_call(override_pystac_version: str) -> None:
    cat = TestCases.case_1()
    d = cat.to_dict()
    assert d["stac_version"] == override_pystac_version



import json
import logging
from copy import deepcopy
from typing import Any, cast

import pytest

import pystac.errors
from pystac import Asset, Collection, Item, ItemAssetDefinition
from pystac.errors import STACError
from pystac.extensions.classification import Classification
from pystac.extensions.mlm import (
    ARCHITECTURE_PROP,
    NAME_PROP,
    TASKS_PROP,
    AcceleratorType,
    AssetDetailedMLMExtension,
    AssetGeneralMLMExtension,
    Hyperparameters,
    InputStructure,
    ItemMLMExtension,
    MLMExtension,
    ModelBand,
    ModelInput,
    ModelOutput,
    ProcessingExpression,
    ResizeType,
    ResultStructure,
    TaskType,
    ValueScaling,
    ValueScalingType,
)
from pystac.extensions.raster import DataType
from tests.utils import TestCases

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()

BASIC_MLM_ITEM_URI = TestCases.get_path("data-files/mlm/item_basic.json")
PLAIN_ITEM_URI = TestCases.get_path("data-files/item/sample-item.json")
MLM_COLLECTION_URI = TestCases.get_path("data-files/mlm/collection.json")


@pytest.fixture
def basic_item_dict() -> dict[str, Any]:
    with open(BASIC_MLM_ITEM_URI) as f:
        return cast(dict[str, Any], json.load(f))


@pytest.fixture
def basic_mlm_item() -> Item:
    return Item.from_file(BASIC_MLM_ITEM_URI)


@pytest.fixture
def plain_item() -> Item:
    return Item.from_file(PLAIN_ITEM_URI)


@pytest.fixture
def mlm_collection() -> Collection:
    return Collection.from_file(MLM_COLLECTION_URI)


def test_stac_extension(basic_mlm_item: Item) -> None:
    assert MLMExtension.has_extension(basic_mlm_item)


def test_model_band() -> None:
    d = {"name": "asdf", "format": "qwer", "expression": "asdf"}
    c = ModelBand.create(**d)
    assert c.name == d["name"]
    assert c.format == d["format"]
    assert c.expression == d["expression"]

    assert c.to_dict() == d

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_model_props() -> None:
    c = ModelBand({})
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.name

    c.name = "asdf"
    assert c.name == "asdf"

    c.format = "asdf"
    assert c.name == "asdf"

    assert c.expression is None
    c.expression = "asdf"
    assert c.expression == "asdf"


def test_processing_expression() -> None:
    d = {"format": "python", "expression": "asdf"}
    c = ProcessingExpression.create(**d)
    assert c.format == d["format"]
    assert c.expression == d["expression"]

    assert c.to_dict() == d

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_processint_expression_props() -> None:
    c = ProcessingExpression({})

    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.format
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.expression

    c.format = "python"
    assert c.format == "python"

    c.expression = "B01 + B02"
    assert c.expression == "B01 + B02"


@pytest.mark.parametrize(
    "scale_type, min_val, max_val, mean, stddev, value, format_val, expression",
    [
        (ValueScalingType.MIN_MAX, 0, 4, 3, 3, 4, "asdf", "asdf"),
        (ValueScalingType.MIN_MAX, 0.2, 4.3, 3.13, 3.2, 4.5, "asdf", "asdf"),
        (ValueScalingType.MIN_MAX, 0, 4, None, None, None, None, None),
        (ValueScalingType.SCALE, None, None, None, None, 2, None, None),
    ],
)
def test_valuescaling_object(
    scale_type: ValueScalingType,
    min_val: int | float | None,
    max_val: int | float | None,
    mean: int | float | None,
    stddev: int | float | None,
    value: int | float | None,
    format_val: str | None,
    expression: str | None,
) -> None:
    c = ValueScaling.create(
        scale_type,
        minimum=min_val,
        maximum=max_val,
        mean=mean,
        stddev=stddev,
        value=value,
        format=format_val,
        expression=expression,
    )
    assert c.type == scale_type
    assert c.minimum == min_val
    assert c.maximum == max_val
    assert c.mean == mean
    assert c.stddev == stddev
    assert c.value == value
    assert c.format == format_val
    assert c.expression == expression

    with pytest.raises(STACError):
        ValueScaling.create(
            ValueScalingType.MIN_MAX, minimum=1
        )  # missing param maximum

    with pytest.raises(STACError):
        ValueScaling.create(ValueScalingType.Z_SCORE, mean=3)  # missing param stddev

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_valuescaling_required_params() -> None:
    assert ValueScaling.get_required_props(ValueScalingType.MIN_MAX) == [
        "minimum",
        "maximum",
    ]
    assert ValueScaling.get_required_props(ValueScalingType.Z_SCORE) == [
        "mean",
        "stddev",
    ]
    assert ValueScaling.get_required_props(ValueScalingType.CLIP) == [
        "minimum",
        "maximum",
    ]
    assert ValueScaling.get_required_props(ValueScalingType.CLIP_MIN) == ["minimum"]
    assert ValueScaling.get_required_props(ValueScalingType.CLIP_MAX) == ["maximum"]
    assert ValueScaling.get_required_props(ValueScalingType.OFFSET) == ["value"]
    assert ValueScaling.get_required_props(ValueScalingType.SCALE) == ["value"]
    assert ValueScaling.get_required_props(ValueScalingType.PROCESSING) == [
        "format",
        "expression",
    ]


def test_input_structure() -> None:
    c = InputStructure.create(
        shape=[-1, 3, 64, 64],
        dim_order=["batch", "channel", "width", "height"],
        data_type=DataType.FLOAT64,
    )
    assert c.shape == [-1, 3, 64, 64]
    assert c.dim_order == ["batch", "channel", "width", "height"]
    assert c.data_type == DataType.FLOAT64

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_model_input_structure_props() -> None:
    c = InputStructure({})

    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.shape
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.dim_order
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.data_type

    c.shape = [1]
    assert c.shape == [1]

    c.dim_order = ["bands"]
    assert c.dim_order == ["bands"]

    c.data_type = DataType.FLOAT64
    assert c.data_type == DataType.FLOAT64


input_testdata = [
    (
        ["B02", "B03", "B04"],
        ValueScaling.create(ValueScalingType.SCALE, value=3),
        ResizeType.CROP,
        ProcessingExpression.create("python", "asdf"),
    ),
    (
        ["B02", "B03", "B04"],
        ValueScaling.create(ValueScalingType.SCALE, value=3),
        None,
        ProcessingExpression.create("python", "asdf"),
    ),
    (
        [ModelBand.create("B02"), ModelBand.create("B03"), ModelBand.create("B04")],
        ValueScaling.create(ValueScalingType.SCALE, value=3),
        ResizeType.CROP,
        ProcessingExpression.create("python", "asdf"),
    ),
    (
        ["B02", "B03", "B04"],
        None,
        ResizeType.CROP,
        ProcessingExpression.create("python", "asdf"),
    ),
    (
        ["B02", "B03", "B04"],
        ValueScaling.create(ValueScalingType.SCALE, value=3),
        ResizeType.CROP,
        None,
    ),
]


@pytest.mark.parametrize(
    "bands, value_scaling, resize_type, pre_processing_function", input_testdata
)
def test_model_input(
    bands: list[str] | list[ModelBand],
    value_scaling: ValueScaling | None,
    resize_type: ResizeType | None,
    pre_processing_function: ProcessingExpression | None,
) -> None:
    input_structure = InputStructure.create(
        [1, 2, 3], ["batch", "width", "length"], DataType.FLOAT64
    )

    c = ModelInput.create(
        name="asdf",
        bands=bands,
        input=input_structure,
        description="foo",
        value_scaling=value_scaling,
        resize_type=resize_type,
        pre_processing_function=pre_processing_function,
    )
    assert c.name == "asdf"
    assert c.bands == bands
    assert c.input == input_structure
    assert c.description == "foo"
    assert c.value_scaling == value_scaling
    assert c.resize_type == resize_type
    assert c.pre_processing_function == pre_processing_function

    # assert that some attributes are always included in dict, even if they are None
    d_reverse = c.to_dict()
    assert "value_scaling" in d_reverse
    assert "resize_type" in d_reverse
    assert "pre_processing_function" in d_reverse

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_model_input_props() -> None:
    c = ModelInput({})
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.name
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.bands
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.input

    c.name = "asdf"
    assert c.name == "asdf"

    c.bands = ["B02", "B03"]
    assert c.bands == ["B02", "B03"]

    inp = InputStructure.create([12], ["bands"], DataType.FLOAT64)
    c.input = inp
    assert c.input == inp

    assert c.value_scaling is None
    val_obj = ValueScaling.create(ValueScalingType.SCALE, value=3)
    c.value_scaling = val_obj
    assert c.value_scaling == val_obj

    assert c.resize_type is None
    c.resize_type = ResizeType.CROP
    assert c.resize_type == ResizeType.CROP

    assert c.pre_processing_function is None
    exp = ProcessingExpression.create("python", "asdf")
    c.pre_processing_function = exp
    assert c.pre_processing_function == exp


def test_result_structure() -> None:
    c = ResultStructure.create(
        shape=[1, 64, 64],
        dim_order=["time", "width", "height"],
        data_type=DataType.FLOAT64,
    )
    assert c.shape == [1, 64, 64]
    assert c.dim_order == ["time", "width", "height"]
    assert c.data_type == DataType.FLOAT64

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_result_structure_props() -> None:
    c = ResultStructure({})

    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.shape
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.dim_order
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.data_type

    c.shape = [1, 2, 3]
    c.dim_order = ["batch", "band", "time"]
    c.data_type = DataType.FLOAT64

    assert c.shape == [1, 2, 3]
    assert c.dim_order == ["batch", "band", "time"]
    assert c.data_type == DataType.FLOAT64


@pytest.mark.parametrize(
    "post_proc_func", (ProcessingExpression.create("asdf", "asdf"), None)
)
def test_model_output(post_proc_func: ProcessingExpression | None) -> None:
    c = ModelOutput.create(
        name="asdf",
        tasks=[TaskType.DETECTION, TaskType.OBJECT_DETECTION],
        result=ResultStructure.create([1, 2, 3], ["a", "b", "c"], DataType.FLOAT64),
        description="asdf",
        classes=[
            Classification.create(1, name="a"),
            Classification.create(2, name="b"),
        ],
        post_processing_function=post_proc_func,
    )
    assert c.name == "asdf"
    assert c.tasks == [TaskType.DETECTION, TaskType.OBJECT_DETECTION]
    assert c.result == ResultStructure.create(
        [1, 2, 3], ["a", "b", "c"], DataType.FLOAT64
    )
    assert c.description == "asdf"
    assert c.classes == [
        Classification.create(1, name="a"),
        Classification.create(2, name="b"),
    ]
    assert c.post_processing_function == post_proc_func

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def test_model_output_props() -> None:
    c = ModelOutput({})

    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.name
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.tasks
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = c.result

    c.name = "asdf"
    assert c.name == "asdf"

    c.tasks = [TaskType.CLASSIFICATION]
    assert c.tasks == [TaskType.CLASSIFICATION]

    res = ResultStructure.create([1], ["band"], DataType.FLOAT64)
    c.result = res
    assert c.result == res

    assert c.description is None
    c.description = "asdf"
    assert c.description == "asdf"

    assert c.classes is None
    c.classes = [Classification.create(value=3, name="foo")]
    assert c.classes == [Classification.create(value=3, name="foo")]

    exp = ProcessingExpression.create("python", "asdf")
    assert c.post_processing_function is None
    c.post_processing_function = exp
    assert c.post_processing_function == exp


def test_hyperparameters() -> None:
    d = {
        "nms_max_detections": 500,
        "nms_threshold": 0.25,
        "iou_threshold": 0.5,
        "random_state": 12345,
    }
    c = Hyperparameters.create(**d)
    for key in d:
        assert key in c.to_dict()
        assert c.to_dict()[key] == d[key]

    with pytest.raises(NotImplementedError):
        _ = c == "blah"


def teest_get_schema_uri(basic_mlm_item: Item) -> None:
    with pytest.raises(DeprecationWarning):
        assert any(
            [
                uri in basic_mlm_item.stac_extensions
                for uri in MLMExtension.get_schema_uris()
            ]
        )


def test_ext_raises_if_item_does_not_conform(plain_item: Item) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        MLMExtension.ext(plain_item)


@pytest.mark.vcr()
def test_apply(plain_item: Item) -> None:
    plain_item.assets["model"] = Asset(
        href="https://google.com",
        roles=["mlm:model"],
        extra_fields={"mlm:artifact_type": "asdf"},
    )

    MLMExtension.add_to(plain_item)
    assert MLMExtension.get_schema_uri() in plain_item.stac_extensions

    model_input = [
        ModelInput.create(
            name="modelinput",
            bands=[],
            input=InputStructure.create(
                shape=[-1, 64, 64],
                dim_order=["batch", "width", "height"],
                data_type=DataType.FLOAT64,
            ),
        )
    ]
    model_output = [
        ModelOutput.create(
            name="modeloutput",
            tasks=[TaskType.CLASSIFICATION],
            result=ResultStructure.create(
                shape=[1, 64, 64],
                dim_order=["batch", "width", "height"],
                data_type=DataType.FLOAT64,
            ),
        )
    ]
    hyp = Hyperparameters.create(
        nms_max_detections=500,
        nms_threshold=0.25,
        iou_threshold=0.5,
        random_state=12345,
    )

    MLMExtension.ext(plain_item).apply(
        name="asdf",
        architecture="ResNet",
        tasks=[TaskType.CLASSIFICATION],
        framework="PyTorch",
        framework_version="1.2.3",
        memory_size=3,
        total_parameters=123,
        pretrained=True,
        pretrained_source="asdfasdfasdf",
        batch_size_suggestion=32,
        accelerator=AcceleratorType.CUDA,
        accelerator_constrained=False,
        accelerator_summary="This is the summary",
        accelerator_count=1,
        input=model_input,
        output=model_output,
        hyperparameters=hyp,
    )
    plain_item.validate()

    assert (
        MLMExtension.ext(plain_item).mlm_name is not None
        and MLMExtension.ext(plain_item).mlm_name == "asdf"
    )

    assert (
        MLMExtension.ext(plain_item).architecture is not None
        and MLMExtension.ext(plain_item).architecture == "ResNet"
    )

    assert MLMExtension.ext(plain_item).tasks is not None and MLMExtension.ext(
        plain_item
    ).tasks == [TaskType.CLASSIFICATION]

    assert (
        MLMExtension.ext(plain_item).framework is not None
        and MLMExtension.ext(plain_item).framework == "PyTorch"
    )

    assert (
        MLMExtension.ext(plain_item).framework_version is not None
        and MLMExtension.ext(plain_item).framework_version == "1.2.3"
    )

    assert (
        MLMExtension.ext(plain_item).memory_size is not None
        and MLMExtension.ext(plain_item).memory_size == 3
    )

    assert (
        MLMExtension.ext(plain_item).total_parameters is not None
        and MLMExtension.ext(plain_item).total_parameters == 123
    )

    assert (
        MLMExtension.ext(plain_item).pretrained is not None
        and MLMExtension.ext(plain_item).pretrained is True
    )

    assert (
        MLMExtension.ext(plain_item).pretrained_source is not None
        and MLMExtension.ext(plain_item).pretrained_source == "asdfasdfasdf"
    )

    assert (
        MLMExtension.ext(plain_item).batch_size_suggestion is not None
        and MLMExtension.ext(plain_item).batch_size_suggestion == 32
    )

    assert (
        MLMExtension.ext(plain_item).accelerator is not None
        and MLMExtension.ext(plain_item).accelerator == AcceleratorType.CUDA
    )

    assert (
        MLMExtension.ext(plain_item).accelerator_constrained is not None
        and MLMExtension.ext(plain_item).accelerator_constrained is False
    )

    assert (
        MLMExtension.ext(plain_item).accelerator_summary is not None
        and MLMExtension.ext(plain_item).accelerator_summary == "This is the summary"
    )

    assert (
        MLMExtension.ext(plain_item).accelerator_count is not None
        and MLMExtension.ext(plain_item).accelerator_count == 1
    )

    assert MLMExtension.ext(plain_item).input is not None and len(
        MLMExtension.ext(plain_item).input
    ) == len(model_input)

    assert MLMExtension.ext(plain_item).input[0] == model_input[0]

    assert MLMExtension.ext(plain_item).output is not None and len(
        MLMExtension.ext(plain_item).output
    ) == len(model_output)

    assert MLMExtension.ext(plain_item).output[0] == model_output[0]

    assert (
        MLMExtension.ext(plain_item).hyperparameters is not None
        and MLMExtension.ext(plain_item).hyperparameters == hyp
    )

    d = {
        **plain_item.properties,
        "mlm:name": "asdf",
        "mlm:architecture": "ResNet",
        "mlm:tasks": [TaskType.CLASSIFICATION],
        "mlm:framework": "PyTorch",
        "mlm:framework_version": "1.2.3",
        "mlm:memory_size": 3,
        "mlm:total_parameters": 123,
        "mlm:pretrained": True,
        "mlm:pretrained_source": "asdfasdfasdf",
        "mlm:batch_size_suggestion": 32,
        "mlm:accelerator": AcceleratorType.CUDA,
        "mlm:accelerator_constrained": False,
        "mlm:accelerator_summary": "This is the summary",
        "mlm:accelerator_count": 1,
        "mlm:input": [inp.to_dict() for inp in model_input],
        "mlm:output": [out.to_dict() for out in model_output],
        "mlm:hyperparameters": hyp.to_dict(),
    }

    assert MLMExtension.ext(plain_item).to_dict() == d


def test_apply_wrong_object() -> None:
    with pytest.raises(pystac.ExtensionTypeError):
        _ = MLMExtension.ext(1, False)


def test_to_from_dict(basic_item_dict: dict[str, Any]) -> None:
    d1 = deepcopy(basic_item_dict)
    d2 = Item.from_dict(basic_item_dict, migrate=False).to_dict()

    assert d1 == d2


def test_add_to_item(plain_item: Item) -> None:
    # check that URI gets added
    assert MLMExtension.get_schema_uri() not in plain_item.stac_extensions
    MLMExtension.add_to(plain_item)
    assert MLMExtension.get_schema_uri() in plain_item.stac_extensions

    # Assure that it gets added only once
    MLMExtension.add_to(plain_item)
    MLMExtension.add_to(plain_item)

    mlm_uris = [
        uri
        for uri in plain_item.stac_extensions
        if uri == MLMExtension.get_schema_uri()
    ]
    assert len(mlm_uris) == 1


@pytest.mark.vcr()
def test_validate_mlm(basic_mlm_item: Item) -> None:
    basic_mlm_item.validate()


def test_add_required_props(plain_item: Item) -> None:
    item_ext = MLMExtension.ext(plain_item, add_if_missing=True)

    assert isinstance(item_ext, ItemMLMExtension)

    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = item_ext.mlm_name
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = item_ext.architecture
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = item_ext.tasks
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = item_ext.input
    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        _ = item_ext.output

    model_input = ModelInput.create(
        name="DummyModel",
        bands=["B04", "B08"],
        input=InputStructure.create(
            shape=[2, 10], dim_order=["time", "bands"], data_type=DataType.FLOAT64
        ),
    )
    model_output = ModelOutput.create(
        name="out",
        tasks=[TaskType.CLASSIFICATION],
        result=ResultStructure.create(
            shape=[1], dim_order=["classification"], data_type=DataType.FLOAT64
        ),
    )
    item_ext.mlm_name = "DummyModel"
    item_ext.architecture = "ResNet"
    item_ext.tasks = [TaskType.CLASSIFICATION]
    item_ext.input = [model_input]
    item_ext.output = [model_output]

    assert item_ext.mlm_name == "DummyModel"
    assert item_ext.architecture == "ResNet"
    assert item_ext.tasks == [TaskType.CLASSIFICATION]
    assert item_ext.input == [model_input]
    assert item_ext.output == [model_output]


def test_add_optional_props(plain_item: Item) -> None:
    item_ext = MLMExtension.ext(plain_item, add_if_missing=True)

    assert isinstance(item_ext, ItemMLMExtension)
    assert item_ext.framework is None
    item_ext.framework = "pytorch"
    assert item_ext.framework == "pytorch"

    assert item_ext.framework_version is None
    item_ext.framework_version = "1.0.0"
    assert item_ext.framework_version == "1.0.0"

    assert item_ext.memory_size is None
    item_ext.memory_size = 3
    assert item_ext.memory_size == 3

    assert item_ext.total_parameters is None
    item_ext.total_parameters = 10000
    assert item_ext.total_parameters == 10000

    assert item_ext.pretrained is None
    item_ext.pretrained = True
    assert item_ext.pretrained is True

    assert item_ext.pretrained_source is None
    item_ext.pretrained_source = "asdf"
    assert item_ext.pretrained_source == "asdf"

    assert item_ext.batch_size_suggestion is None
    item_ext.batch_size_suggestion = 64
    assert item_ext.batch_size_suggestion == 64

    assert item_ext.accelerator is None
    item_ext.accelerator = AcceleratorType.CUDA
    assert item_ext.accelerator == AcceleratorType.CUDA

    assert item_ext.accelerator_constrained is None
    item_ext.accelerator_constrained = False
    assert item_ext.accelerator_constrained is False

    assert item_ext.accelerator_summary is None
    item_ext.accelerator_summary = "Summary"
    assert item_ext.accelerator_summary == "Summary"

    assert item_ext.accelerator_count is None
    item_ext.accelerator_count = 1
    assert item_ext.accelerator_count == 1


def test_add_to_asset(plain_item: Item) -> None:
    MLMExtension.ext(plain_item, add_if_missing=True)
    asset = plain_item.assets["analytic"]

    assert NAME_PROP not in asset.extra_fields.keys()
    assert ARCHITECTURE_PROP not in asset.extra_fields.keys()
    assert TASKS_PROP not in asset.extra_fields.keys()

    asset_ext = AssetDetailedMLMExtension.ext(asset)
    asset_ext.mlm_name = "asdf"
    asset_ext.architecture = "ResNet"
    asset_ext.tasks = [TaskType.CLASSIFICATION]

    assert NAME_PROP in asset.extra_fields.keys()
    assert ARCHITECTURE_PROP in asset.extra_fields.keys()
    assert TASKS_PROP in asset.extra_fields.keys()

    assert asset.extra_fields[NAME_PROP] == "asdf"
    assert asset.extra_fields[ARCHITECTURE_PROP] == "ResNet"
    assert asset.extra_fields[TASKS_PROP] == [TaskType.CLASSIFICATION]


@pytest.mark.parametrize("is_model_asset", (True, False))
def test_asset_props(plain_item: Item, is_model_asset: bool) -> None:
    asset = plain_item.assets["analytic"]
    asset_ext = AssetGeneralMLMExtension.ext(asset, add_if_missing=True)

    assert asset_ext.artifact_type is None
    assert asset_ext.compile_method is None
    assert asset_ext.entrypoint is None

    # test special behavior if the asset has role "mlm:model"
    if is_model_asset and isinstance(asset_ext.asset.roles, list):
        asset_ext.asset.roles.append("mlm:model")
        with pytest.raises(pystac.errors.RequiredPropertyMissing):
            _ = asset_ext.artifact_type

    asset_ext.artifact_type = "foo"
    asset_ext.compile_method = "bar"
    asset_ext.entrypoint = "baz"

    assert asset_ext.artifact_type == "foo"
    assert asset_ext.compile_method == "bar"
    assert asset_ext.entrypoint == "baz"


def test_add_to_generic_asset() -> None:
    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
        extra_fields={
            "mlm:artifact_type": "foo",
            "mlm:compile_method": "bar",
            "mlm:entrypoint": "baz",
        },
    )
    asset_ext = AssetGeneralMLMExtension.ext(asset, add_if_missing=False)
    assert asset_ext.artifact_type == "foo"
    assert asset_ext.compile_method == "bar"
    assert asset_ext.entrypoint == "baz"


def test_apply_generic_asset() -> None:
    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
    )
    asset_ext = AssetGeneralMLMExtension.ext(asset, add_if_missing=False)
    asset_ext.apply(artifact_type="foo", compile_method="bar", entrypoint="baz")
    assert asset_ext.artifact_type == "foo"
    assert asset_ext.compile_method == "bar"
    assert asset_ext.entrypoint == "baz"


def test_to_dict_asset_generic() -> None:
    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
    )
    asset_ext = AssetGeneralMLMExtension.ext(asset, add_if_missing=False)
    asset_ext.apply(artifact_type="foo", compile_method="bar", entrypoint="baz")

    d = {
        "mlm:artifact_type": "foo",
        "mlm:compile_method": "bar",
        "mlm:entrypoint": "baz",
    }
    assert asset_ext.to_dict() == d


def test_add_to_detailled_asset() -> None:
    model_input = ModelInput.create(
        name="model",
        bands=["B02"],
        input=InputStructure.create(
            shape=[1], dim_order=["batch"], data_type=DataType.FLOAT64
        ),
    )
    model_output = ModelOutput.create(
        name="output",
        tasks=[TaskType.CLASSIFICATION],
        result=ResultStructure.create(
            shape=[1], dim_order=["batch"], data_type=DataType.FLOAT64
        ),
    )

    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
        extra_fields={
            "mlm:name": "asdf",
            "mlm:architecture": "ResNet",
            "mlm:tasks": [TaskType.CLASSIFICATION],
            "mlm:input": [model_input.to_dict()],
            "mlm:output": [model_output.to_dict()],
            "mlm:artifact_type": "foo",
            "mlm:compile_method": "bar",
            "mlm:entrypoint": "baz",
        },
    )

    asset_ext = AssetDetailedMLMExtension.ext(asset, add_if_missing=False)

    assert asset_ext.mlm_name == "asdf"
    assert asset_ext.architecture == "ResNet"
    assert asset_ext.tasks == [TaskType.CLASSIFICATION]
    assert asset_ext.input == [model_input]
    assert asset_ext.output == [model_output]
    assert asset_ext.artifact_type == "foo"
    assert asset_ext.compile_method == "bar"
    assert asset_ext.entrypoint == "baz"

    assert repr(asset_ext) == f"<AssetDetailedMLMExtension Asset href={asset.href}>"

    with pytest.raises(pystac.errors.RequiredPropertyMissing):
        asset_ext.artifact_type = None

    asset_ext.compile_method = None
    assert asset_ext.compile_method is None

    asset_ext.entrypoint = None
    assert asset_ext.entrypoint is None

    asset.roles.remove("mlm:model") if isinstance(asset.roles, list) else None
    asset_ext.artifact_type = None
    assert asset_ext.artifact_type is None


def test_correct_asset_extension_is_used() -> None:
    asset = Asset("https://example.com")
    assert isinstance(asset.ext.mlm, AssetGeneralMLMExtension)

    asset.extra_fields["mlm:name"] = "asdf"
    assert isinstance(asset.ext.mlm, AssetDetailedMLMExtension)


def test_asset_accessor() -> None:
    item_asset = ItemAssetDefinition.create(
        title="asdf", description="asdf", media_type="asdf", roles=["asdf"]
    )
    assert isinstance(item_asset.ext.mlm, MLMExtension)


def test_apply_detailled_asset() -> None:
    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
    )
    asset_ext = AssetDetailedMLMExtension.ext(asset, add_if_missing=False)

    model_input = ModelInput.create(
        name="model",
        bands=["B02"],
        input=InputStructure.create(
            shape=[1], dim_order=["batch"], data_type=DataType.FLOAT64
        ),
    )
    model_output = ModelOutput.create(
        name="output",
        tasks=[TaskType.CLASSIFICATION],
        result=ResultStructure.create(
            shape=[1], dim_order=["batch"], data_type=DataType.FLOAT64
        ),
    )

    asset_ext.apply(
        "asdf",
        "ResNet",
        [TaskType.CLASSIFICATION],
        [model_input],
        [model_output],
        artifact_type="foo",
        compile_method="bar",
        entrypoint="baz",
    )

    assert asset_ext.mlm_name == "asdf"
    assert asset_ext.architecture == "ResNet"
    assert asset_ext.tasks == [TaskType.CLASSIFICATION]
    assert asset_ext.input == [model_input]
    assert asset_ext.output == [model_output]
    assert asset_ext.artifact_type == "foo"
    assert asset_ext.compile_method == "bar"
    assert asset_ext.entrypoint == "baz"


def test_to_dict_detailed_asset() -> None:
    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
    )
    asset_ext = AssetDetailedMLMExtension.ext(asset, add_if_missing=False)

    model_input = ModelInput.create(
        name="model",
        bands=["B02"],
        input=InputStructure.create(
            shape=[1], dim_order=["batch"], data_type=DataType.FLOAT64
        ),
    )
    model_output = ModelOutput.create(
        name="output",
        tasks=[TaskType.CLASSIFICATION],
        result=ResultStructure.create(
            shape=[1], dim_order=["batch"], data_type=DataType.FLOAT64
        ),
    )

    asset_ext.apply(
        "asdf",
        "ResNet",
        [TaskType.CLASSIFICATION],
        [model_input],
        [model_output],
        artifact_type="foo",
        compile_method="bar",
        entrypoint="baz",
    )

    d = {
        "mlm:name": "asdf",
        "mlm:architecture": "ResNet",
        "mlm:tasks": [TaskType.CLASSIFICATION],
        "mlm:input": [model_input.to_dict()],
        "mlm:output": [model_output.to_dict()],
        "mlm:artifact_type": "foo",
        "mlm:compile_method": "bar",
        "mlm:entrypoint": "baz",
        "mlm:accelerator": None,
        "mlm:pretrained_source": None,
    }
    assert asset_ext.to_dict() == d


def test_item_asset_extension(mlm_collection: Collection) -> None:
    assert mlm_collection.item_assets
    item_asset = mlm_collection.item_assets["weights"]
    item_asset_ext = MLMExtension.ext(item_asset, add_if_missing=True)
    assert MLMExtension.get_schema_uri() in mlm_collection.stac_extensions
    assert mlm_collection.item_assets["weights"].ext.has("mlm")

    assert (
        repr(item_asset_ext)
        == f"<ItemAssetsMLMExtension ItemAssetDefinition={item_asset}"
    )


def test_collection_extension(mlm_collection: Collection) -> None:
    coll_ext = MLMExtension.ext(mlm_collection, add_if_missing=True)
    assert MLMExtension.get_schema_uri() in mlm_collection.stac_extensions
    assert mlm_collection.ext.has("mlm")

    coll_ext.mlm_name = "asdf"
    assert coll_ext.mlm_name == "asdf"
    assert (
        repr(coll_ext) == f"<CollectionMLMExtension Collection id={mlm_collection.id}>"
    )


def test_raise_exception_on_mlm_extension_and_asset() -> None:
    asset = pystac.Asset(
        href="http://example.com/test.tiff",
        title="image",
        description="asdf",
        media_type="application/tiff",
        roles=["mlm:model"],
    )
    with pytest.raises(TypeError):
        MLMExtension.ext(asset, add_if_missing=False)



"""Implements the :stac-ext:`Datacube Extension <datacube>`."""

from __future__ import annotations

from abc import ABC
from collections.abc import Iterable
from typing import Any, Generic, Literal, TypeVar, cast

import pystac
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import StringEnum, get_required, map_opt

#: Generalized version of :class:`~pystac.Collection`, `:class:`~pystac.Item`,
#: :class:`~pystac.Asset`, or :class:`~pystac.ItemAssetDefinition`
T = TypeVar(
    "T", pystac.Collection, pystac.Item, pystac.Asset, pystac.ItemAssetDefinition
)

SCHEMA_URI = "https://stac-extensions.github.io/datacube/v2.2.0/schema.json"

PREFIX: str = "cube:"
DIMENSIONS_PROP = PREFIX + "dimensions"
VARIABLES_PROP = PREFIX + "variables"

# Dimension properties
DIM_TYPE_PROP = "type"
DIM_DESC_PROP = "description"
DIM_AXIS_PROP = "axis"
DIM_EXTENT_PROP = "extent"
DIM_VALUES_PROP = "values"
DIM_STEP_PROP = "step"
DIM_REF_SYS_PROP = "reference_system"
DIM_UNIT_PROP = "unit"

# Variable properties
VAR_TYPE_PROP = "type"
VAR_DESC_PROP = "description"
VAR_EXTENT_PROP = "extent"
VAR_VALUES_PROP = "values"
VAR_DIMENSIONS_PROP = "dimensions"
VAR_UNIT_PROP = "unit"


class DimensionType(StringEnum):
    """Dimension object types for spatial and temporal Dimension Objects."""

    SPATIAL = "spatial"
    GEOMETRIES = "geometries"
    TEMPORAL = "temporal"


class HorizontalSpatialDimensionAxis(StringEnum):
    """Allowed values for ``axis`` field of :class:`HorizontalSpatialDimension`
    object."""

    X = "x"
    Y = "y"


class VerticalSpatialDimensionAxis(StringEnum):
    """Allowed values for ``axis`` field of :class:`VerticalSpatialDimension`
    object."""

    Z = "z"


class Dimension(ABC):
    """Object representing a dimension of the datacube. The fields contained in
    Dimension Object vary by ``type``. See the :stac-ext:`Datacube Dimension Object
    <datacube#dimension-object>` docs for details.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    @property
    def dim_type(self) -> DimensionType | str:
        """The type of the dimension. Must be ``"spatial"`` for
        :stac-ext:`Horizontal Spatial Dimension Objects
        <datacube#horizontal-spatial-raster-dimension-object>` or
        :stac-ext:`Vertical Spatial Dimension Objects
        <datacube#vertical-spatial-dimension-object>`, ``geometries`` for
        :stac-ext:`Spatial Vector Dimension Objects
        <datacube#spatial-vector-dimension-object>` ``"temporal"`` for
        :stac-ext:`Temporal Dimension Objects
        <datacube#temporal-dimension-object>`. May be an arbitrary string for
        :stac-ext:`Additional Dimension Objects
        <datacube#additional-dimension-object>`."""
        return cast(
            str,
            get_required(
                self.properties.get(DIM_TYPE_PROP), "cube:dimension", DIM_TYPE_PROP
            ),
        )

    @dim_type.setter
    def dim_type(self, v: DimensionType | str) -> None:
        self.properties[DIM_TYPE_PROP] = v

    @property
    def description(self) -> str | None:
        """Detailed multi-line description to explain the dimension. `CommonMark 0.29
        <http://commonmark.org/>`__ syntax MAY be used for rich text representation."""
        return self.properties.get(DIM_DESC_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(DIM_DESC_PROP, None)
        else:
            self.properties[DIM_DESC_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        return self.properties

    @staticmethod
    def from_dict(d: dict[str, Any]) -> Dimension:
        dim_type: str = get_required(
            d.get(DIM_TYPE_PROP), "cube_dimension", DIM_TYPE_PROP
        )
        if dim_type == DimensionType.SPATIAL:
            axis: str = get_required(
                d.get(DIM_AXIS_PROP), "cube_dimension", DIM_AXIS_PROP
            )
            if axis == "z":
                return VerticalSpatialDimension(d)
            else:
                return HorizontalSpatialDimension(d)
        elif dim_type == DimensionType.GEOMETRIES:
            return VectorSpatialDimension(d)
        elif dim_type == DimensionType.TEMPORAL:
            # The v1.0.0 spec says that AdditionalDimensions can have
            # type 'temporal', but it is unclear how to differentiate that
            # from a temporal dimension. Just key off of type for now.
            # See https://github.com/stac-extensions/datacube/issues/5
            return TemporalDimension(d)
        else:
            return AdditionalDimension(d)


class SpatialDimension(Dimension):
    @property
    def extent(self) -> list[float] | None:
        """Extent (lower and upper bounds) of the dimension as two-dimensional array.
        Open intervals with ``None`` are not allowed."""
        return cast(
            list[float],
            get_required(
                self.properties.get(DIM_EXTENT_PROP), "cube:dimension", DIM_EXTENT_PROP
            ),
        )

    @extent.setter
    def extent(self, v: list[float] | None) -> None:
        self.properties[DIM_EXTENT_PROP] = v

    @property
    def values(self) -> list[float] | None:
        """Optional set of all potential values."""
        return self.properties.get(DIM_VALUES_PROP)

    @values.setter
    def values(self, v: list[float] | None) -> None:
        if v is None:
            self.properties.pop(DIM_VALUES_PROP, None)
        else:
            self.properties[DIM_VALUES_PROP] = v

    @property
    def step(self) -> float | None:
        """The space between the values. Use ``None`` for irregularly spaced steps."""
        return self.properties.get(DIM_STEP_PROP)

    @step.setter
    def step(self, v: float | None) -> None:
        self.properties[DIM_STEP_PROP] = v

    def clear_step(self) -> None:
        """Setting step to None sets it to the null value,
        which means irregularly spaced steps. Use clear_step
        to remove it from the properties."""
        self.properties.pop(DIM_STEP_PROP, None)

    @property
    def reference_system(self) -> str | float | dict[str, Any] | None:
        """The spatial reference system for the data, specified as `numerical EPSG code
        <http://www.epsg-registry.org/>`__, `WKT2 (ISO 19162) string
        <http://docs.opengeospatial.org/is/18-010r7/18-010r7.html>`__ or `PROJJSON
        object <https://proj.org/specifications/projjson.html>`__.
        Defaults to EPSG code 4326."""
        return self.properties.get(DIM_REF_SYS_PROP)

    @reference_system.setter
    def reference_system(self, v: str | float | dict[str, Any] | None) -> None:
        if v is None:
            self.properties.pop(DIM_REF_SYS_PROP, None)
        else:
            self.properties[DIM_REF_SYS_PROP] = v


class HorizontalSpatialDimension(SpatialDimension):
    @property
    def axis(self) -> HorizontalSpatialDimensionAxis:
        """Axis of the spatial dimension. Must be one of ``"x"`` or ``"y"``."""
        return cast(
            HorizontalSpatialDimensionAxis,
            get_required(
                self.properties.get(DIM_AXIS_PROP), "cube:dimension", DIM_AXIS_PROP
            ),
        )

    @axis.setter
    def axis(self, v: HorizontalSpatialDimensionAxis) -> None:
        self.properties[DIM_AXIS_PROP] = v


class VerticalSpatialDimension(SpatialDimension):
    @property
    def axis(self) -> VerticalSpatialDimensionAxis:
        """Axis of the spatial dimension. Must be ``"z"``."""
        return cast(
            VerticalSpatialDimensionAxis,
            get_required(
                self.properties.get(DIM_AXIS_PROP), "cube:dimension", DIM_AXIS_PROP
            ),
        )

    @axis.setter
    def axis(self, v: VerticalSpatialDimensionAxis) -> None:
        self.properties[DIM_AXIS_PROP] = v

    @property
    def extent(self) -> list[float] | None:
        """Extent (lower and upper bounds) of the dimension as two-dimensional array.
        Open intervals with ``None`` are not allowed."""
        return self.properties.get(DIM_EXTENT_PROP)

    @extent.setter
    def extent(self, v: list[float] | None) -> None:
        if v is None:
            self.properties.pop(DIM_EXTENT_PROP, None)
        else:
            self.properties[DIM_EXTENT_PROP] = v

    @property
    def unit(self) -> str | None:
        """The unit of measurement for the data, preferably compliant to `UDUNITS-2
        <https://ncics.org/portfolio/other-resources/udunits2/>`__ units (singular)."""
        return self.properties.get(DIM_UNIT_PROP)

    @unit.setter
    def unit(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(DIM_UNIT_PROP, None)
        else:
            self.properties[DIM_UNIT_PROP] = v


class VectorSpatialDimension(Dimension):
    @property
    def axes(self) -> list[str] | None:
        """Axes of the vector dimension as an ordered set of `x`, `y` and `z`."""
        return self.properties.get("axes")

    @axes.setter
    def axes(self, v: list[str]) -> None:
        if v is None:
            self.properties.pop("axes", None)
        else:
            self.properties["axes"] = v

    @property
    def bbox(self) -> list[float]:
        """A single bounding box of the geometries as defined for STAC
        Collections but not nested."""
        return get_required(self.properties.get("bbox"), "cube:bbox", "bbox")

    @bbox.setter
    def bbox(self, v: list[float]) -> None:
        self.properties["bbox"] = v

    @property
    def values(self) -> list[str] | None:
        """Optionally, a representation of the geometries. This could be a list
        of WKT strings or other identifiers."""
        return self.properties.get(DIM_VALUES_PROP)

    @values.setter
    def values(self, v: list[str] | None) -> None:
        if v is None:
            self.properties.pop(DIM_VALUES_PROP, None)
        else:
            self.properties[DIM_VALUES_PROP] = v

    @property
    def geometry_types(self) -> list[str] | None:
        """A set of geometry types. If not present, mixed geometry types must be
        assumed."""
        return self.properties.get("geometry_types")

    @geometry_types.setter
    def geometry_types(self, v: list[str] | None) -> None:
        if v is None:
            self.properties.pop("geometry_types", None)
        else:
            self.properties["geometry_types"] = v

    @property
    def reference_system(self) -> str | float | dict[str, Any] | None:
        """The reference system for the data."""
        return self.properties.get(DIM_REF_SYS_PROP)

    @reference_system.setter
    def reference_system(self, v: str | float | dict[str, Any] | None) -> None:
        if v is None:
            self.properties.pop(DIM_REF_SYS_PROP, None)
        else:
            self.properties[DIM_REF_SYS_PROP] = v


class TemporalDimension(Dimension):
    @property
    def extent(self) -> list[str | None] | None:
        """Extent (lower and upper bounds) of the dimension as two-dimensional array.
        The dates and/or times must be strings compliant to `ISO 8601
        <https://en.wikipedia.org/wiki/ISO_8601>`__. ``None`` is allowed for open date
        ranges."""
        return self.properties.get(DIM_EXTENT_PROP)

    @extent.setter
    def extent(self, v: list[str | None] | None) -> None:
        if v is None:
            self.properties.pop(DIM_EXTENT_PROP, None)
        else:
            self.properties[DIM_EXTENT_PROP] = v

    @property
    def values(self) -> list[str] | None:
        """If the dimension consists of set of specific values they can be listed here.
        The dates and/or times must be strings compliant to `ISO 8601
        <https://en.wikipedia.org/wiki/ISO_8601>`__."""
        return self.properties.get(DIM_VALUES_PROP)

    @values.setter
    def values(self, v: list[str] | None) -> None:
        if v is None:
            self.properties.pop(DIM_VALUES_PROP, None)
        else:
            self.properties[DIM_VALUES_PROP] = v

    @property
    def step(self) -> str | None:
        """The space between the temporal instances as `ISO 8601 duration
        <https://en.wikipedia.org/wiki/ISO_8601#Durations>`__, e.g. P1D. Use null for
        irregularly spaced steps."""
        return self.properties.get(DIM_STEP_PROP)

    @step.setter
    def step(self, v: str | None) -> None:
        self.properties[DIM_STEP_PROP] = v

    def clear_step(self) -> None:
        """Setting step to None sets it to the null value,
        which means irregularly spaced steps. Use clear_step
        to remove it from the properties."""
        self.properties.pop(DIM_STEP_PROP, None)


class AdditionalDimension(Dimension):
    @property
    def extent(self) -> list[float | None] | None:
        """If the dimension consists of `ordinal
        <https://en.wikipedia.org/wiki/Level_of_measurement#Ordinal_scale>`__ values,
        the extent (lower and upper bounds) of the values as two-dimensional array. Use
        null for open intervals."""
        return self.properties.get(DIM_EXTENT_PROP)

    @extent.setter
    def extent(self, v: list[float | None] | None) -> None:
        if v is None:
            self.properties.pop(DIM_EXTENT_PROP, None)
        else:
            self.properties[DIM_EXTENT_PROP] = v

    @property
    def values(self) -> list[str] | list[float] | None:
        """A set of all potential values, especially useful for `nominal
        <https://en.wikipedia.org/wiki/Level_of_measurement#Nominal_level>`__ values."""
        return self.properties.get(DIM_VALUES_PROP)

    @values.setter
    def values(self, v: list[str] | list[float] | None) -> None:
        if v is None:
            self.properties.pop(DIM_VALUES_PROP, None)
        else:
            self.properties[DIM_VALUES_PROP] = v

    @property
    def step(self) -> float | None:
        """If the dimension consists of `interval
        <https://en.wikipedia.org/wiki/Level_of_measurement#Interval_scale>`__ values,
        the space between the values. Use null for irregularly spaced steps."""
        return self.properties.get(DIM_STEP_PROP)

    @step.setter
    def step(self, v: float | None) -> None:
        self.properties[DIM_STEP_PROP] = v

    def clear_step(self) -> None:
        """Setting step to None sets it to the null value,
        which means irregularly spaced steps. Use clear_step
        to remove it from the properties."""
        self.properties.pop(DIM_STEP_PROP, None)

    @property
    def unit(self) -> str | None:
        """The unit of measurement for the data, preferably compliant to `UDUNITS-2
        units <https://ncics.org/portfolio/other-resources/udunits2/>`__ (singular)."""
        return self.properties.get(DIM_UNIT_PROP)

    @unit.setter
    def unit(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(DIM_UNIT_PROP, None)
        else:
            self.properties[DIM_UNIT_PROP] = v

    @property
    def reference_system(self) -> str | float | dict[str, Any] | None:
        """The reference system for the data."""
        return self.properties.get(DIM_REF_SYS_PROP)

    @reference_system.setter
    def reference_system(self, v: str | float | dict[str, Any] | None) -> None:
        if v is None:
            self.properties.pop(DIM_REF_SYS_PROP, None)
        else:
            self.properties[DIM_REF_SYS_PROP] = v


class VariableType(StringEnum):
    """Variable object types"""

    DATA = "data"
    AUXILIARY = "auxiliary"


class Variable:
    """Object representing a variable in the datacube. The dimensions field lists
    zero or more :stac-ext:`Datacube Dimension Object <datacube#dimension-object>`
    instances. See the :stac-ext:`Datacube Variable Object
    <datacube#variable-object>` docs for details.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    @property
    def dimensions(self) -> list[str]:
        """The dimensions of the variable. Should refer to keys in the
        ``cube:dimensions`` object or be an empty list if the variable has no
        dimensions
        """
        return get_required(
            self.properties.get(VAR_DIMENSIONS_PROP),
            "cube:variable",
            VAR_DIMENSIONS_PROP,
        )

    @dimensions.setter
    def dimensions(self, v: list[str]) -> None:
        self.properties[VAR_DIMENSIONS_PROP] = v

    @property
    def var_type(self) -> VariableType | str:
        """Type of the variable, either ``data`` or ``auxiliary``"""
        return cast(
            str,
            get_required(
                self.properties.get(VAR_TYPE_PROP), "cube:variable", VAR_TYPE_PROP
            ),
        )

    @var_type.setter
    def var_type(self, v: VariableType | str) -> None:
        self.properties[VAR_TYPE_PROP] = v

    @property
    def description(self) -> str | None:
        """Detailed multi-line description to explain the variable. `CommonMark 0.29
        <http://commonmark.org/>`__ syntax MAY be used for rich text representation."""
        return self.properties.get(VAR_DESC_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(VAR_DESC_PROP, None)
        else:
            self.properties[VAR_DESC_PROP] = v

    @property
    def extent(self) -> list[float | str | None]:
        """If the variable consists of `ordinal values
        <https://en.wikipedia.org/wiki/Level_of_measurement#Ordinal_scale>`, the extent
        (lower and upper bounds) of the values as two-dimensional array. Use ``None``
        for open intervals"""
        return get_required(
            self.properties.get(VAR_EXTENT_PROP), "cube:variable", VAR_EXTENT_PROP
        )

    @extent.setter
    def extent(self, v: list[float | str | None]) -> None:
        self.properties[VAR_EXTENT_PROP] = v

    @property
    def values(self) -> list[float | str] | None:
        """A set of all potential values, especially useful for `nominal values
        <https://en.wikipedia.org/wiki/Level_of_measurement#Nominal_level>`."""
        return self.properties.get(VAR_VALUES_PROP)

    @values.setter
    def values(self, v: list[float | str] | None) -> None:
        if v is None:
            self.properties.pop(VAR_VALUES_PROP)
        else:
            self.properties[VAR_VALUES_PROP] = v

    @property
    def unit(self) -> str | None:
        """The unit of measurement for the data, preferably compliant to `UDUNITS-2
        <https://ncics.org/portfolio/other-resources/udunits2/>` units (singular)"""
        return self.properties.get(VAR_UNIT_PROP)

    @unit.setter
    def unit(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(VAR_UNIT_PROP)
        else:
            self.properties[VAR_UNIT_PROP] = v

    @staticmethod
    def from_dict(d: dict[str, Any]) -> Variable:
        return Variable(d)

    def to_dict(self) -> dict[str, Any]:
        return self.properties


class DatacubeExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of a
    :class:`~pystac.Collection`, :class:`~pystac.Item`, or :class:`~pystac.Asset` with
    properties from the :stac-ext:`Datacube Extension <datacube>`. This class is
    generic over the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`DatacubeExtension`, use the
    :meth:`DatacubeExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> dc_ext = DatacubeExtension.ext(item)
    """

    name: Literal["cube"] = "cube"

    def apply(
        self,
        dimensions: dict[str, Dimension],
        variables: dict[str, Variable] | None = None,
    ) -> None:
        """Applies Datacube Extension properties to the extended
        :class:`~pystac.Collection`, :class:`~pystac.Item` or :class:`~pystac.Asset`.

        Args:
            dimensions : Dictionary mapping dimension name to :class:`Dimension`
                objects.
            variables : Dictionary mapping variable name to a :class:`Variable`
                object.
        """
        self.dimensions = dimensions
        self.variables = variables

    @property
    def dimensions(self) -> dict[str, Dimension]:
        """A dictionary where each key is the name of a dimension and each
        value is a :class:`~Dimension` object.
        """
        result = get_required(
            self._get_property(DIMENSIONS_PROP, dict[str, Any]), self, DIMENSIONS_PROP
        )
        return {k: Dimension.from_dict(v) for k, v in result.items()}

    @dimensions.setter
    def dimensions(self, v: dict[str, Dimension]) -> None:
        self._set_property(DIMENSIONS_PROP, {k: dim.to_dict() for k, dim in v.items()})

    @property
    def variables(self) -> dict[str, Variable] | None:
        """A dictionary where each key is the name of a variable and each
        value is a :class:`~Variable` object.
        """
        result = self._get_property(VARIABLES_PROP, dict[str, Any])

        if result is None:
            return None
        return {k: Variable.from_dict(v) for k, v in result.items()}

    @variables.setter
    def variables(self, v: dict[str, Variable] | None) -> None:
        self._set_property(
            VARIABLES_PROP,
            map_opt(
                lambda variables: {k: var.to_dict() for k, var in variables.items()}, v
            ),
        )

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> DatacubeExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Datacube
        Extension <datacube>`.

        This extension can be applied to instances of :class:`~pystac.Collection`,
        :class:`~pystac.Item` or :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(DatacubeExtension[T], CollectionDatacubeExtension(obj))
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(DatacubeExtension[T], ItemDatacubeExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(DatacubeExtension[T], AssetDatacubeExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(DatacubeExtension[T], ItemAssetsDatacubeExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class CollectionDatacubeExtension(DatacubeExtension[pystac.Collection]):
    """A concrete implementation of :class:`DatacubeExtension` on an
    :class:`~pystac.Collection` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Datacube Extension <datacube>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`DatacubeExtension.ext` on an :class:`~pystac.Collection` to extend it.
    """

    collection: pystac.Collection
    properties: dict[str, Any]

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields

    def __repr__(self) -> str:
        return f"<CollectionDatacubeExtension Item id={self.collection.id}>"


class ItemDatacubeExtension(DatacubeExtension[pystac.Item]):
    """A concrete implementation of :class:`DatacubeExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Datacube Extension <datacube>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`DatacubeExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    properties: dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemDatacubeExtension Item id={self.item.id}>"


class AssetDatacubeExtension(DatacubeExtension[pystac.Asset]):
    """A concrete implementation of :class:`DatacubeExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`Datacube Extension <datacube>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`DatacubeExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    properties: dict[str, Any]
    additional_read_properties: Iterable[dict[str, Any]] | None

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]
        else:
            self.additional_read_properties = None

    def __repr__(self) -> str:
        return f"<AssetDatacubeExtension Item id={self.asset_href}>"


class ItemAssetsDatacubeExtension(DatacubeExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class DatacubeExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "datacube",
        "https://stac-extensions.github.io/datacube/v1.0.0/schema.json",
        "https://stac-extensions.github.io/datacube/v2.0.0/schema.json",
        "https://stac-extensions.github.io/datacube/v2.1.0/schema.json",
    }
    stac_object_types = {
        pystac.STACObjectType.COLLECTION,
        pystac.STACObjectType.ITEM,
    }


DATACUBE_EXTENSION_HOOKS: ExtensionHooks = DatacubeExtensionHooks()
